from typing import Any, Dict, List, Optional, cast


def str_or_none(v: Any) -> Optional[str]:
    return str(v) if v is not None else None


def mapping_or_none(value: object) -> Optional[Dict[str, Any]]:
    if not isinstance(value, dict):
        return None
    return cast("Dict[str, Any]", value)  # pragma: allow-cast yaml mapping typed narrowing


def list_or_none(value: object) -> Optional[List[Any]]:
    if not isinstance(value, list):
        return None
    return cast("List[Any]", value)  # pragma: allow-cast yaml list typed narrowing


__all__ = []



import logging
from typing import Any, Dict, Hashable, List, Tuple, cast

from .....events import EVENT_FIELD_COMPUTE
from .....planning.operators import ComputeOperatorIr, SupportedOperatorIr
from .....secure_compute_contracts import is_secure_compute_calculator
from .....spec.ir import ComputeCallContextIr, DerivedFieldIr
from .....vendor.compact.typing_extensionsx import override
from ....context import BatchContext
from ...runtime.runtime import ExecutionRuntime
from ..base import OperatorExecutor
from .errors import handle_compute_error
from .payloads import build_field_compute_dependencies_payload

_EXPECTED_COMPUTE_ERRORS = (
    TypeError,
    ValueError,
    ZeroDivisionError,
    ArithmeticError,
)


def _execute_constant_compute(
    field_spec: "DerivedFieldIr",
    context: BatchContext,
    batch_row_nth: List[Hashable],
    runtime: ExecutionRuntime,
    *,
    compute_mode: str,
    wants_field_compute: bool,
) -> None:
    constant_dep_values: Dict[str, Any] = {}
    try:
        result = field_spec.compute(**constant_dep_values)
    except _EXPECTED_COMPUTE_ERRORS as exc:  # type: ignore[misc]
        for row_id in batch_row_nth:
            handle_compute_error(
                runtime,
                context,
                field_key=field_spec.field_id,
                row_id=row_id,
                dependencies=constant_dep_values,
                dependency_names=field_spec.dependencies,
                exc=exc,
                compute_mode=compute_mode,
                unexpected=False,
            )
        return
    except Exception as exc:
        first_row_id = batch_row_nth[0] if batch_row_nth else None
        logging.exception(  # noqa: LOG015
            "字段计算发生未预期的异常: 字段=%s, 行标识=%s",
            field_spec.field_id,
            first_row_id,
        )
        for row_id in batch_row_nth:
            handle_compute_error(
                runtime,
                context,
                field_key=field_spec.field_id,
                row_id=row_id,
                dependencies=constant_dep_values,
                dependency_names=field_spec.dependencies,
                exc=exc,
                compute_mode=compute_mode,
                unexpected=True,
            )
        return

    for row_id in batch_row_nth:
        context.set_field_value(field_spec.field_id, row_id, result)
        if wants_field_compute:
            runtime.instrumentation.emit_field_compute(field_spec.field_id, row_id, constant_dep_values, result)


def _execute_secure_compute(
    field_spec: "DerivedFieldIr",
    context: BatchContext,
    batch_row_nth: List[Hashable],
    runtime: ExecutionRuntime,
    *,
    compute_mode: str,
    wants_field_compute: bool,
) -> None:
    deps = field_spec.dependencies
    calculator = field_spec.calculator
    value_formatter = field_spec.value_formatter
    guardrails_enabled = runtime.guardrails.enabled

    for row_id in batch_row_nth:
        dep_args: Tuple[Any, ...] = tuple(context.get_field_value(dep_key, row_id) for dep_key in deps)

        try:
            result = calculator(*dep_args)
            if value_formatter is not None:
                result = value_formatter(result)
            context.set_field_value(field_spec.field_id, row_id, result)

            if wants_field_compute:
                dep_values_payload = build_field_compute_dependencies_payload(deps, dep_args)
                runtime.instrumentation.emit_field_compute(field_spec.field_id, row_id, dep_values_payload, result)
        except _EXPECTED_COMPUTE_ERRORS as exc:  # type: ignore[misc]
            deps_payload: Dict[str, Any] = {}
            if not guardrails_enabled:
                deps_payload = build_field_compute_dependencies_payload(deps, dep_args)
            handle_compute_error(
                runtime,
                context,
                field_key=field_spec.field_id,
                row_id=row_id,
                dependencies=deps_payload,
                dependency_names=deps,
                exc=exc,
                compute_mode=compute_mode,
                unexpected=False,
            )
        except Exception as exc:
            logging.exception(  # noqa: LOG015
                "字段计算发生未预期的异常: 字段=%s, 行标识=%s",
                field_spec.field_id,
                row_id,
            )
            deps_payload = {}
            if not guardrails_enabled:
                deps_payload = build_field_compute_dependencies_payload(deps, dep_args)
            handle_compute_error(
                runtime,
                context,
                field_key=field_spec.field_id,
                row_id=row_id,
                dependencies=deps_payload,
                dependency_names=deps,
                exc=exc,
                compute_mode=compute_mode,
                unexpected=True,
            )


def _execute_general_compute(
    field_spec: Any,
    context: BatchContext,
    batch_row_nth: List[Hashable],
    runtime: ExecutionRuntime,
    *,
    compute_mode: str,
    wants_field_compute: bool,
) -> None:
    for row_id in batch_row_nth:
        dep_values: Dict[str, Any] = {}
        for dep_key in field_spec.dependencies:
            dep_values[dep_key] = context.get_field_value(dep_key, row_id)

        try:
            ctx_key = field_spec.call_ctx_key if isinstance(field_spec, DerivedFieldIr) else None
            if ctx_key:
                ctx_value = ComputeCallContextIr(
                    row_id=row_id,
                    batch_num=runtime.batch_num,
                    field_id=field_spec.field_id,
                    deps=field_spec.dependencies,
                    values=dep_values,
                )
                if wants_field_compute:
                    compute_kwargs = dict(dep_values)
                    compute_kwargs[ctx_key] = ctx_value
                    result = field_spec.compute(**compute_kwargs)
                else:
                    dep_values[ctx_key] = ctx_value
                    try:
                        result = field_spec.compute(**dep_values)
                    finally:
                        del dep_values[ctx_key]
            else:
                result = field_spec.compute(**dep_values)

            context.set_field_value(field_spec.field_id, row_id, result)
            if wants_field_compute:
                runtime.instrumentation.emit_field_compute(field_spec.field_id, row_id, dep_values, result)
        except _EXPECTED_COMPUTE_ERRORS as exc:  # type: ignore[misc]
            handle_compute_error(
                runtime,
                context,
                field_key=field_spec.field_id,
                row_id=row_id,
                dependencies=dep_values,
                dependency_names=field_spec.dependencies,
                exc=exc,
                compute_mode=compute_mode,
                unexpected=False,
            )
        except Exception as exc:
            logging.exception(  # noqa: LOG015
                "字段计算发生未预期的异常: 字段=%s, 行标识=%s",
                field_spec.field_id,
                row_id,
            )
            handle_compute_error(
                runtime,
                context,
                field_key=field_spec.field_id,
                row_id=row_id,
                dependencies=dep_values,
                dependency_names=field_spec.dependencies,
                exc=exc,
                compute_mode=compute_mode,
                unexpected=True,
            )


class ComputeOperatorExecutor(OperatorExecutor):
    """计算算子执行器."""

    @override
    def execute(
        self,
        operator: SupportedOperatorIr,
        context: BatchContext,
        batch_row_nth: List[Hashable],
        runtime: ExecutionRuntime,
    ) -> None:
        op = cast("ComputeOperatorIr", operator)  # pragma: allow-cast operator dispatch typed narrowing
        field_spec = op.field_spec
        guardrails = runtime.guardrails
        compute_mode = guardrails.effective_compute_mode()
        wants_field_compute = runtime.instrumentation.wants(EVENT_FIELD_COMPUTE)

        if isinstance(field_spec, DerivedFieldIr) and field_spec.is_constant_compute:
            _execute_constant_compute(
                field_spec,
                context,
                batch_row_nth,
                runtime,
                compute_mode=compute_mode,
                wants_field_compute=wants_field_compute,
            )
            return

        if isinstance(field_spec, DerivedFieldIr) and not field_spec.call_ctx_key and is_secure_compute_calculator(field_spec.calculator):
            _execute_secure_compute(
                field_spec,
                context,
                batch_row_nth,
                runtime,
                compute_mode=compute_mode,
                wants_field_compute=wants_field_compute,
            )
            return

        _execute_general_compute(
            field_spec,
            context,
            batch_row_nth,
            runtime,
            compute_mode=compute_mode,
            wants_field_compute=wants_field_compute,
        )


__all__ = [
    "ComputeOperatorExecutor",
]



from collections.abc import Set as AbstractSet
from typing import Any, Optional, Set, Tuple

from ..._internal.loggingx import prefix
from ...events import (
    EVENT_ADAPTIVE_SCHEDULER_DECISION,
    EVENT_BATCH_END,
    EVENT_BATCH_START,
    EVENT_COLUMN_WRITE,
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_ERROR,
    EVENT_FIELD_COMPUTE,
    EVENT_FIELD_SLIM,
    EVENT_LOADER_CALL,
    EVENT_LOADER_RETRY,
    EVENT_LOADER_SLIM,
    EVENT_OUTPUT_TARGET_END,
    EVENT_PIPELINE_END,
    EVENT_PIPELINE_START,
    EVENT_RELATION_LOOKUP,
    EVENT_ROW_RELEASE,
    EVENT_ROW_WRITE,
    EVENT_STAGE_SPAN,
    EVENT_WORKFLOW_CACHE_ACQUIRE,
    EVENT_WORKFLOW_CACHE_EVICT,
    EVENT_WORKFLOW_CACHE_RELEASE,
    EVENT_WORKFLOW_NODE_CANCELLED,
    EVENT_WORKFLOW_NODE_END,
    EVENT_WORKFLOW_NODE_START,
    EVENT_WORKFLOW_RESOURCE_COMMIT,
    EVENT_WORKFLOW_RESOURCE_CREATE,
    EVENT_WORKFLOW_RESOURCE_DISCARD,
    EVENT_WORKFLOW_RESOURCE_WRITE,
)
from ...exceptions import ScalimObserverError

OBSERVER_RAISED_EXCEPTION_WARNING = prefix("ob") + "观察者 %s.%s 抛出异常"
OBSERVER_CLOSE_RAISED_EXCEPTION_WARNING = prefix("ob") + "观察者 %s 关闭时抛出异常"

CATALOG_EVENT_TYPES: Tuple[str, ...] = (
    EVENT_PIPELINE_START,
    EVENT_PIPELINE_END,
    EVENT_BATCH_START,
    EVENT_BATCH_END,
    EVENT_LOADER_CALL,
    EVENT_LOADER_RETRY,
    EVENT_FIELD_COMPUTE,
    EVENT_ERROR,
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_FIELD_SLIM,
    EVENT_ROW_WRITE,
    EVENT_ROW_RELEASE,
    EVENT_LOADER_SLIM,
    EVENT_COLUMN_WRITE,
    EVENT_RELATION_LOOKUP,
    EVENT_STAGE_SPAN,
    EVENT_ADAPTIVE_SCHEDULER_DECISION,
    EVENT_OUTPUT_TARGET_END,
    EVENT_WORKFLOW_NODE_START,
    EVENT_WORKFLOW_NODE_END,
    EVENT_WORKFLOW_NODE_CANCELLED,
    EVENT_WORKFLOW_CACHE_ACQUIRE,
    EVENT_WORKFLOW_CACHE_RELEASE,
    EVENT_WORKFLOW_CACHE_EVICT,
    EVENT_WORKFLOW_RESOURCE_CREATE,
    EVENT_WORKFLOW_RESOURCE_WRITE,
    EVENT_WORKFLOW_RESOURCE_COMMIT,
    EVENT_WORKFLOW_RESOURCE_DISCARD,
)

CATALOG_EVENT_TYPES_SET: Set[str] = set(CATALOG_EVENT_TYPES)
DEFAULT_MAX_RECORDED_EVENTS = 10_000
CAPTURE_OVERFLOW_POLICIES = ("raise", "drop-oldest", "drop-newest")


class ScalimObserverCaptureOverflowError(ScalimObserverError):
    pass


def validate_event_types(observer: Any, value: Any) -> Optional[Set[str]]:
    if value is None:
        return None
    if not isinstance(value, AbstractSet):
        msg = "observer.event_types must be None or Set[str]; got {} for {}".format(type(value).__name__, type(observer).__name__)
        raise TypeError(msg)
    normalized: Set[str] = set()
    for item in value:
        if not isinstance(item, str):
            msg = "observer.event_types must contain only str; got {} element {!r} for {}".format(
                type(item).__name__,
                item,
                type(observer).__name__,
            )
            raise TypeError(msg)
        normalized.add(item)
    return normalized


__all__ = []



"""`workflow` 共享输出资源: `sheetbook` 实现(内部模块).

说明:
- 承载 `sheetbook_sheet`/`sheetbook_append` 的内存物化、读取与导出
- 运行时需兼容 `Python 3.6`
"""

from abc import ABC
from contextlib import suppress
from pathlib import Path
from typing import Any, Dict, FrozenSet, Iterator, List, Optional, Tuple, cast

from .._internal.utils.excel import escape_excel_formula
from ..events import EVENT_DIAGNOSTIC_WARNING
from ..events._events import DiagnosticWarningEvent
from ..sinks._internal.base import create_temp_path
from ..vendor.compact.typing_extensionsx import override
from ..vendor.dataclassesx import dataclass
from .resources_base import ScalimWorkflowWriteError, WorkflowResourceManagerBase, acquire_write_lock, release_write_lock
from .resources_csv import WorkflowCsvInput, build_alignment_mapping, describe_header_diff, iter_csv_rows, read_csv_header
from .resources_workbook import best_effort_close_write_only_workbook_worksheets, get_openpyxl_workbook_class

# 内部实现仍沿用原有局部命名,减少重构噪音.
_build_alignment_mapping = build_alignment_mapping
_describe_header_diff = describe_header_diff
_iter_csv_rows = iter_csv_rows
_read_csv_header = read_csv_header
_best_effort_close_write_only_workbook_worksheets = best_effort_close_write_only_workbook_worksheets
_get_openpyxl_workbook_class = get_openpyxl_workbook_class


def _materialize_aligned_csv_rows(expected: List[str], mapping: List[int], *, input_csv: WorkflowCsvInput) -> List[List[str]]:
    _ = list(expected)
    rows: List[List[str]] = []
    for row in _iter_csv_rows(input_csv):
        out_row: List[str] = []
        for src_idx in mapping:
            out_row.append(row[src_idx] if src_idx >= 0 and src_idx < len(row) else "")
        rows.append(out_row)
    return rows


def _sorted_sheetbook_segments(segments: List["_SheetBookSegment"]) -> List["_SheetBookSegment"]:
    return sorted(segments, key=lambda seg: (int(seg.decl_order), str(seg.producer_node_id)))


def _write_sheetbook_plan_to_openpyxl_workbook(workbook: object, plan: "_SheetBookPlan") -> None:
    wb = cast("Any", workbook)  # pragma: allow-cast openpyxl workbook runtime boundary
    for sheet_name in plan.sheet_order:
        sheet_plan = plan.sheets.get(sheet_name)
        if sheet_plan is None:
            continue  # pragma: no cover  # pragma: allow-no-cover unreachable: sheet_plan always exists
        ws = wb.create_sheet(str(sheet_name))
        header_written = False
        fields = list(sheet_plan.baseline_header)
        escaped_fields = [escape_excel_formula(x, allow_formulas=bool(plan.export_allow_formulas)) for x in fields]
        for seg in _sorted_sheetbook_segments(sheet_plan.segments):
            if seg.header_policy == "always" or (seg.header_policy == "once" and not header_written):
                _ = ws.append(list(escaped_fields))
                header_written = True
            # `never`: 不输出表头

            for row in seg.rows:
                out_row = [escape_excel_formula(v, allow_formulas=bool(plan.export_allow_formulas)) for v in row]
                _ = ws.append(out_row)


def _save_openpyxl_workbook_atomic(workbook: object, *, output_path: str) -> None:
    wb = cast("Any", workbook)  # pragma: allow-cast openpyxl workbook runtime boundary
    temp_path = create_temp_path(output_path, ".xlsx.tmp")
    temp_obj = Path(temp_path)
    try:
        wb.save(temp_obj)
        _ = temp_obj.replace(output_path)
    except Exception as exc:
        with suppress(Exception):
            temp_obj.unlink()
        msg = "Sheetbook export failed: {}: {}".format(type(exc).__name__, exc)
        raise ScalimWorkflowWriteError(msg) from exc


@dataclass(frozen=True)
class SheetBookDef:
    resource_id: str
    budget_max_sheets: int
    budget_max_total_cells: int
    export_path: Optional[str]
    export_write_lock: bool
    export_allow_formulas: bool = False


@dataclass
class _SheetBookSegment:
    producer_node_id: str
    decl_order: int
    rows: List[List[str]]
    header_policy: str


@dataclass
class _SheetBookSheetPlan:
    sheet: str
    baseline_header: List[str]
    segments: List[_SheetBookSegment]
    cell_count: int = 0


@dataclass
class _SheetBookPlan:
    resource_id: str
    budget_max_sheets: int
    budget_max_total_cells: int
    export_path: Optional[str]
    export_write_lock: bool
    export_allow_formulas: bool
    sheet_decl_order: Dict[str, int]
    sheet_order: List[str]
    sheets: Dict[str, _SheetBookSheetPlan]
    lock_path: Optional[Path] = None
    total_cells: int = 0
    last_workflow_node_id: Optional[str] = None


class _WorkflowSheetBookResourceMixin(WorkflowResourceManagerBase, ABC):
    @staticmethod
    def _sheetbook_has_alignment_mismatch(expected: List[str], input_header: List[str], *, align_by: str) -> bool:
        if str(align_by) == "field_id":
            exp_set = {str(x) for x in expected}
            act_set = {str(x) for x in input_header}
            missing = exp_set.difference(act_set)
            extra = act_set.difference(exp_set)
            return bool(missing or extra)
        return list(input_header) != list(expected)

    def _emit_sheetbook_commit(self, plan: _SheetBookPlan, *, display_path: str) -> None:
        node_id = plan.last_workflow_node_id or "__wf__commit"
        self._emit_resource_commit(
            workflow_node_id=node_id, resource_type="sheetbook", resource_id=plan.resource_id, path=str(display_path)
        )

    def _maybe_acquire_sheetbook_write_lock(self, plan: _SheetBookPlan, *, output_path: str) -> None:
        if not bool(plan.export_write_lock):
            return
        lock_path = acquire_write_lock(
            output_path,
            owner={
                "workflow_exec_id": self._workflow_exec_id,
                "resource_type": "sheetbook",
                "resource_id": str(plan.resource_id),
            },
        )
        plan.lock_path = lock_path

    def _release_sheetbook_write_lock(self, plan: _SheetBookPlan) -> None:
        if plan.lock_path is None:
            return
        release_write_lock(plan.lock_path)
        plan.lock_path = None

    def _sheetbook_sheet_prepare_action(
        self,
        plan: _SheetBookPlan,
        *,
        workflow_node_id: str,
        sheetbook_id: str,
        sheet_name: str,
        on_conflict: str,
    ) -> Tuple[str, bool]:
        action = "write"
        pending_skip = False
        with self._lock:
            existing = plan.sheets.get(sheet_name)
            if existing is not None:
                if on_conflict == "skip":
                    action = "skip"
                    plan.last_workflow_node_id = str(workflow_node_id)
                    pending_skip = True
                if on_conflict == "error":
                    msg = "Sheet conflict (sheetbook_sheet): sheetbook={!r}, sheet={!r}".format(str(sheetbook_id), sheet_name)
                    raise ScalimWorkflowWriteError(msg, diff=["on_conflict=error", "existing_sheet=present"])
                if on_conflict == "overwrite":
                    action = "overwrite"
            elif len(plan.sheets) >= int(plan.budget_max_sheets):
                msg = "Sheetbook budget exceeded: max_sheets (sheetbook={!r})".format(str(sheetbook_id))
                diff = [
                    "budget.max_sheets={}".format(int(plan.budget_max_sheets)),
                    "current_sheets={}".format(len(plan.sheets)),
                    "new_sheet={!r}".format(sheet_name),
                ]
                raise ScalimWorkflowWriteError(msg, diff=diff)
        return action, pending_skip

    def _sheetbook_sheet_store(
        self,
        plan: _SheetBookPlan,
        *,
        workflow_node_id: str,
        sheet_name: str,
        decl_order: int,
        input_node_id: str,
        expected: List[str],
        rows: List[List[str]],
        new_sheet_cells: int,
    ) -> None:
        with self._lock:
            existing = plan.sheets.get(sheet_name)
            old_cells = 0
            if existing is not None:
                old_cells = int(existing.cell_count)

            new_total = int(plan.total_cells) - int(old_cells) + int(new_sheet_cells)
            self._check_sheetbook_budget(plan, new_total_cells=new_total)

            existing_decl_order = plan.sheet_decl_order.get(sheet_name)
            resolved_decl_order = int(decl_order)
            if existing_decl_order is None or resolved_decl_order < int(existing_decl_order):
                plan.sheet_decl_order[sheet_name] = resolved_decl_order
            plan.sheet_order = sorted(plan.sheet_decl_order.keys(), key=lambda name: (plan.sheet_decl_order.get(name, 0), str(name)))

            sheet_plan = _SheetBookSheetPlan(
                sheet=sheet_name,
                baseline_header=list(expected),
                segments=[],
                cell_count=int(new_sheet_cells),
            )
            sheet_plan.segments.append(
                _SheetBookSegment(
                    producer_node_id=str(input_node_id),
                    decl_order=int(decl_order),
                    rows=rows,
                    header_policy="once",
                )
            )
            plan.sheets[sheet_name] = sheet_plan
            plan.total_cells = int(new_total)
            plan.last_workflow_node_id = str(workflow_node_id)

    def _sheetbook_append_prepare(  # noqa: C901
        self,
        plan: _SheetBookPlan,
        *,
        workflow_node_id: str,
        sheetbook_id: str,
        sheet_name: str,
        decl_order: int,
        input_node_id: str,
        input_header: List[str],
        align_by: str,
        on_mismatch: str,
    ) -> Tuple[List[str], List[int], Optional[DiagnosticWarningEvent], Optional[Dict[str, object]], bool]:
        msg: str
        pending_warning: Optional[DiagnosticWarningEvent] = None
        pending_warning_meta: Optional[Dict[str, object]] = None
        pending_skip = False

        with self._lock:
            sheet_plan = plan.sheets.get(sheet_name)
            if sheet_plan is None:
                if len(plan.sheets) >= int(plan.budget_max_sheets):
                    msg = "Sheetbook budget exceeded: max_sheets (sheetbook={!r})".format(str(sheetbook_id))
                    diff = [
                        "budget.max_sheets={}".format(int(plan.budget_max_sheets)),
                        "current_sheets={}".format(len(plan.sheets)),
                        "new_sheet={!r}".format(sheet_name),
                    ]
                    raise ScalimWorkflowWriteError(msg, diff=diff)
                plan.sheet_decl_order[sheet_name] = int(decl_order)
                plan.sheet_order = sorted(plan.sheet_decl_order.keys(), key=lambda name: (plan.sheet_decl_order.get(name, 0), str(name)))
                sheet_plan = _SheetBookSheetPlan(
                    sheet=sheet_name,
                    baseline_header=list(input_header),
                    segments=[],
                    cell_count=0,
                )
                plan.sheets[sheet_name] = sheet_plan
            else:
                existing_decl_order = plan.sheet_decl_order.get(sheet_name)
                resolved_decl_order = int(decl_order)
                if existing_decl_order is None or resolved_decl_order < int(existing_decl_order):
                    plan.sheet_decl_order[sheet_name] = resolved_decl_order
                    plan.sheet_order = sorted(
                        plan.sheet_decl_order.keys(),
                        key=lambda name: (plan.sheet_decl_order.get(name, 0), str(name)),
                    )

            expected = list(sheet_plan.baseline_header)
            mapping = _build_alignment_mapping(expected, input_header)

            if self._sheetbook_has_alignment_mismatch(expected, input_header, align_by=str(align_by)):
                diff = _describe_header_diff(expected, input_header)
                if on_mismatch == "error":
                    msg = "Field alignment mismatch (sheetbook_append): sheetbook={!r}, sheet={!r}".format(str(sheetbook_id), sheet_name)
                    raise ScalimWorkflowWriteError(msg, diff=diff)
                if on_mismatch == "warn":
                    pending_warning = DiagnosticWarningEvent(
                        message="Field alignment mismatch (warn): sheetbook={!r}, sheet={!r}".format(str(sheetbook_id), sheet_name),
                        source_id=None,
                        field_id=None,
                        lookup_key={"expected": expected, "actual": list(input_header)},
                        row_id=None,
                    )
                    pending_warning_meta = {"workflow_exec_id": self._workflow_exec_id, "workflow_node_id": str(workflow_node_id)}
                if on_mismatch == "skip":
                    plan.last_workflow_node_id = str(workflow_node_id)
                    pending_skip = True

            if not pending_skip:
                # 重复写入检测: 同一生产者不应写入同一工作表多次.
                for seg in sheet_plan.segments:
                    if str(seg.producer_node_id) == str(input_node_id):
                        msg = "Duplicate sheetbook write for the same producer: sheetbook={!r}, sheet={!r}, producer={!r}".format(
                            str(sheetbook_id),
                            sheet_name,
                            str(input_node_id),
                        )
                        raise ScalimWorkflowWriteError(msg, diff=["producer_node_id={!r}".format(str(input_node_id))])
        return expected, mapping, pending_warning, pending_warning_meta, pending_skip

    def _sheetbook_append_apply(
        self,
        plan: _SheetBookPlan,
        *,
        sheetbook_id: str,
        sheet_name: str,
        append_cells: int,
        workflow_node_id: str,
        input_node_id: str,
        decl_order: int,
        rows: List[List[str]],
        header_policy: str,
    ) -> None:
        with self._lock:
            sheet_plan = plan.sheets.get(sheet_name)
            if sheet_plan is None:  # pragma: no cover  # pragma: allow-no-cover unreachable: sheet_plan always exists
                msg = "Sheetbook sheet missing during append: sheetbook={!r}, sheet={!r}".format(str(sheetbook_id), sheet_name)
                raise ScalimWorkflowWriteError(msg)

            new_total = int(plan.total_cells) + int(append_cells)
            self._check_sheetbook_budget(plan, new_total_cells=new_total)
            sheet_plan.segments.append(
                _SheetBookSegment(
                    producer_node_id=str(input_node_id),
                    decl_order=int(decl_order),
                    rows=rows,
                    header_policy=str(header_policy),
                )
            )
            sheet_plan.cell_count = int(sheet_plan.cell_count) + int(append_cells)
            plan.total_cells = int(new_total)
            plan.last_workflow_node_id = str(workflow_node_id)

    def _get_or_create_sheetbook(self, sheetbook_id: str, *, workflow_node_id: str) -> _SheetBookPlan:
        key = str(sheetbook_id)
        with self._lock:
            existing = cast("Optional[_SheetBookPlan]", self._sheetbooks.get(key))  # pragma: allow-cast sheetbook plan typed narrowing
            if existing is not None:
                return existing

            raw_def = self._sheetbook_defs.get(key)
            if raw_def is None:
                msg = "Unknown sheetbook resource id: {!r}".format(key)
                raise ScalimWorkflowWriteError(msg)

            raw_def = cast("SheetBookDef", raw_def)  # pragma: allow-cast sheetbook def typed narrowing
            plan = SheetBookPlan(
                resource_id=str(raw_def.resource_id),
                budget_max_sheets=int(raw_def.budget_max_sheets),
                budget_max_total_cells=int(raw_def.budget_max_total_cells),
                export_path=str(raw_def.export_path) if raw_def.export_path is not None else None,
                export_write_lock=bool(raw_def.export_write_lock),
                export_allow_formulas=bool(raw_def.export_allow_formulas),
                sheet_decl_order={},
                sheet_order=[],
                sheets={},
            )
            self._sheetbooks[key] = plan

        display_path = plan.export_path if plan.export_path is not None else "<memory>"
        self._emit_resource_create(
            workflow_node_id=str(workflow_node_id),
            resource_type="sheetbook",
            resource_id=str(key),
            path=str(display_path),
        )
        return plan

    def _check_sheetbook_budget(
        self,
        plan: _SheetBookPlan,
        *,
        new_total_cells: int,
        allow_over_budget: bool = False,
    ) -> None:
        if allow_over_budget:  # pragma: no cover  # pragma: allow-no-cover test-only budget bypass
            return  # pragma: no cover  # pragma: allow-no-cover test-only budget bypass
        limit = int(plan.budget_max_total_cells)
        if int(new_total_cells) <= limit:
            return
        diff = [
            "budget.max_total_cells={}".format(limit),
            "new_total_cells={}".format(int(new_total_cells)),
        ]
        msg = "Sheetbook budget exceeded: sheetbook={!r}".format(str(plan.resource_id))
        raise ScalimWorkflowWriteError(msg, diff=diff)

    def apply_sheetbook_sheet(
        self,
        *,
        workflow_node_id: str,
        decl_order: int,
        sheetbook_id: str,
        sheet: str,
        input_node_id: str,
        input_output_id: str,
        input_csv: WorkflowCsvInput,
        on_conflict: str,
    ) -> None:
        plan = self._get_or_create_sheetbook(sheetbook_id, workflow_node_id=str(workflow_node_id))
        sheet_name = str(sheet)
        input_header = _read_csv_header(input_csv)
        action, pending_skip = self._sheetbook_sheet_prepare_action(
            plan,
            workflow_node_id=str(workflow_node_id),
            sheetbook_id=str(sheetbook_id),
            sheet_name=sheet_name,
            on_conflict=str(on_conflict),
        )

        if pending_skip:
            display_path = plan.export_path if plan.export_path is not None else "<memory>"
            self._emit_resource_write(
                workflow_node_id=str(workflow_node_id),
                resource_type="sheetbook",
                resource_id=str(sheetbook_id),
                path=str(display_path),
                write_kind="sheetbook_sheet",
                action=action,
                input_node_id=str(input_node_id),
                input_output_id=str(input_output_id),
                sheet=sheet_name,
            )
            return

        expected = list(input_header)
        mapping = _build_alignment_mapping(expected, input_header)

        rows = _materialize_aligned_csv_rows(expected, mapping, input_csv=input_csv)
        row_count = len(rows)
        new_sheet_cells = int(row_count) * len(expected)

        self._sheetbook_sheet_store(
            plan,
            workflow_node_id=str(workflow_node_id),
            sheet_name=sheet_name,
            decl_order=int(decl_order),
            input_node_id=str(input_node_id),
            expected=expected,
            rows=rows,
            new_sheet_cells=int(new_sheet_cells),
        )

        display_path = plan.export_path if plan.export_path is not None else "<memory>"
        self._emit_resource_write(
            workflow_node_id=str(workflow_node_id),
            resource_type="sheetbook",
            resource_id=str(sheetbook_id),
            path=str(display_path),
            write_kind="sheetbook_sheet",
            action=action,
            input_node_id=str(input_node_id),
            input_output_id=str(input_output_id),
            sheet=sheet_name,
        )

    def apply_sheetbook_append(
        self,
        *,
        workflow_node_id: str,
        decl_order: int,
        sheetbook_id: str,
        sheet: str,
        input_node_id: str,
        input_output_id: str,
        input_csv: WorkflowCsvInput,
        align_by: str,
        header_policy: str,
        on_mismatch: str,
    ) -> None:
        plan = self._get_or_create_sheetbook(sheetbook_id, workflow_node_id=str(workflow_node_id))
        sheet_name = str(sheet)
        input_header = _read_csv_header(input_csv)
        expected, mapping, pending_warning, pending_warning_meta, pending_skip = self._sheetbook_append_prepare(
            plan,
            workflow_node_id=str(workflow_node_id),
            sheetbook_id=str(sheetbook_id),
            sheet_name=sheet_name,
            decl_order=int(decl_order),
            input_node_id=str(input_node_id),
            input_header=list(input_header),
            align_by=str(align_by),
            on_mismatch=str(on_mismatch),
        )

        if pending_warning is not None:
            _ = self._instrumentation.emit(EVENT_DIAGNOSTIC_WARNING, pending_warning, meta=pending_warning_meta)

        if pending_skip:
            display_path = plan.export_path if plan.export_path is not None else "<memory>"
            self._emit_resource_write(
                workflow_node_id=str(workflow_node_id),
                resource_type="sheetbook",
                resource_id=str(sheetbook_id),
                path=str(display_path),
                write_kind="sheetbook_append",
                action="skip",
                input_node_id=str(input_node_id),
                input_output_id=str(input_output_id),
                sheet=sheet_name,
            )
            return

        rows = _materialize_aligned_csv_rows(expected, mapping, input_csv=input_csv)

        append_rows = len(rows)
        append_cells = int(append_rows) * len(expected)

        self._sheetbook_append_apply(
            plan,
            sheetbook_id=str(sheetbook_id),
            sheet_name=sheet_name,
            append_cells=int(append_cells),
            workflow_node_id=str(workflow_node_id),
            input_node_id=str(input_node_id),
            decl_order=int(decl_order),
            rows=rows,
            header_policy=str(header_policy),
        )

        display_path = plan.export_path if plan.export_path is not None else "<memory>"
        self._emit_resource_write(
            workflow_node_id=str(workflow_node_id),
            resource_type="sheetbook",
            resource_id=str(sheetbook_id),
            path=str(display_path),
            write_kind="sheetbook_append",
            action="append",
            input_node_id=str(input_node_id),
            input_output_id=str(input_output_id),
            sheet=sheet_name,
        )

    def iter_sheetbook_sheet_rows(  # noqa: C901
        self,
        *,
        consumer_node_id: str,
        visible_producer_node_ids: FrozenSet[str],
        producer_node_id: str,
        sheetbook_id: str,
        sheet: str,
    ) -> Iterator[Dict[str, object]]:
        """读取 `sheetbook` 的行快照(按 `ref.node` 截断;按依赖可见性过滤).

        返回: `Iterator[Dict[str, object]]`, 每行必须为 `JSON-like mapping`.
        """
        _ = str(consumer_node_id)
        producer = str(producer_node_id)
        sb_id = str(sheetbook_id)
        sheet_name = str(sheet)
        visible = frozenset(str(x) for x in visible_producer_node_ids)

        with self._lock:
            plan = cast("Optional[_SheetBookPlan]", self._sheetbooks.get(sb_id))  # pragma: allow-cast sheetbook plan typed narrowing
            if plan is None:
                msg = "Unknown sheetbook resource id: {!r}".format(sb_id)
                raise ValueError(msg)
            sheet_plan = plan.sheets.get(sheet_name)
            if sheet_plan is None:
                msg = "Unknown sheetbook sheet: sheetbook={!r}, sheet={!r}".format(sb_id, sheet_name)
                raise ValueError(msg)

            cutoff_idx = None
            ordered_segments = _sorted_sheetbook_segments(list(sheet_plan.segments))
            for idx, seg in enumerate(ordered_segments):
                if str(seg.producer_node_id) == producer:
                    cutoff_idx = int(idx)
                    break
            if cutoff_idx is None:
                msg = "Unknown sheetbook ref node for sheet: node={!r}, sheetbook={!r}, sheet={!r}".format(producer, sb_id, sheet_name)
                raise ValueError(msg)

            baseline_header = list(sheet_plan.baseline_header)
            segments: List[Tuple[str, List[List[str]]]] = []
            for seg in ordered_segments[: cutoff_idx + 1]:
                seg_producer = str(seg.producer_node_id)
                if seg_producer != producer and seg_producer not in visible:
                    continue
                segments.append((seg_producer, list(seg.rows)))

        def _iter() -> Iterator[Dict[str, object]]:
            for _seg_producer, seg_rows in segments:
                for row_values in seg_rows:
                    row: Dict[str, object] = {}
                    for idx, key in enumerate(baseline_header):
                        row[str(key)] = row_values[idx] if idx >= 0 and idx < len(row_values) else ""
                    yield row

        return _iter()

    @override
    def _commit_sheetbook(self, plan: object) -> None:
        p = cast("_SheetBookPlan", plan)  # pragma: allow-cast sheetbook plan typed narrowing
        export_path = p.export_path
        display_path = export_path if export_path is not None else "<memory>"
        if export_path is None:
            self._emit_sheetbook_commit(p, display_path=str(display_path))
            return

        output_path = str(export_path)
        output_dir = Path(output_path).parent
        output_dir.mkdir(parents=True, exist_ok=True)

        self._maybe_acquire_sheetbook_write_lock(p, output_path=output_path)

        try:
            workbook_cls = _get_openpyxl_workbook_class()
        except ImportError as exc:
            with suppress(Exception):
                self._release_sheetbook_write_lock(p)
            raise ScalimWorkflowWriteError(str(exc)) from exc

        wb = workbook_cls(write_only=True)
        try:
            _write_sheetbook_plan_to_openpyxl_workbook(wb, p)
            _save_openpyxl_workbook_atomic(wb, output_path=output_path)
        except Exception:
            _best_effort_close_write_only_workbook_worksheets(wb)
            raise
        finally:
            with suppress(Exception):
                wb.close()
            with suppress(Exception):
                self._release_sheetbook_write_lock(p)

        self._emit_sheetbook_commit(p, display_path=str(display_path))

    @override
    def _discard_sheetbook(self, plan: object, *, workflow_node_id: str, reason: str) -> None:
        p = cast("_SheetBookPlan", plan)  # pragma: allow-cast sheetbook plan typed narrowing
        with suppress(Exception):
            self._release_sheetbook_write_lock(p)
        node_id = p.last_workflow_node_id or str(workflow_node_id)
        display_path = p.export_path if p.export_path is not None else "<memory>"
        self._emit_resource_discard(
            workflow_node_id=node_id,
            resource_type="sheetbook",
            resource_id=p.resource_id,
            path=str(display_path),
            reason=str(reason),
        )


__all__ = [
    "SheetBookDef",
    "SheetBookPlan",
    "SheetBookSegment",
    "SheetBookSheetPlan",
    "WorkflowSheetBookResourceMixin",
]

SheetBookPlan = _SheetBookPlan
SheetBookSegment = _SheetBookSegment
SheetBookSheetPlan = _SheetBookSheetPlan
WorkflowSheetBookResourceMixin = _WorkflowSheetBookResourceMixin



from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Dict, Mapping, Optional, Sequence, Union, cast

if TYPE_CHECKING:
    from .validator import ConfigValidator

from ...init_var_nodes import parse_init_var_mapping_node
from ...schema_dsl.constants import DEFAULT_BATCH_SIZE, UTF8_ENCODING
from ...schema_dsl.models import (
    BOOK_BUDGET_KEYS,
    BOOK_EXPORT_XLSX_KEYS,
    BOOK_KEYS,
    BOOK_WRITE_DEFAULTS_KEYS,
    DEMAND_KEYS,
    FILE_KEYS,
    OUTPUT_EXTRA_SHEET_KEYS,
    RESOURCES_KEYS,
    BookBudgetConfig,
    BookConfig,
    BookExportXlsxConfig,
    BookWriteDefaultsConfig,
    DemandConfig,
    FileConfig,
    OutputExtraSheetConfig,
    ResourcesConfig,
)
from ...schema_dsl.output_enums import (
    BOOK_KINDS,
    BOOK_WRITE_ALIGN_BY_ENUM,
    BOOK_WRITE_HEADER_POLICY_ENUM,
    BOOK_WRITE_MODE_ENUM,
    BOOK_WRITE_ON_CONFLICT_ENUM,
    BOOK_WRITE_ON_MISMATCH_ENUM,
    DEFAULT_BOOK_WRITE_ALIGN_BY,
    DEFAULT_BOOK_WRITE_HEADER_POLICY,
    DEFAULT_BOOK_WRITE_MODE,
    DEFAULT_BOOK_WRITE_ON_CONFLICT,
    DEFAULT_BOOK_WRITE_ON_MISMATCH,
    FILE_KINDS,
)
from .error_envelope import ErrorEnvelope, ScalimYamlValidationError
from .imports import ScalimYamlImportExpansionError, contains_import_syntax, expand_imports_inplace
from .models import RawDemand
from .parsers.fields import ParserFieldsMixin
from .parsers.guardrails import ParserGuardrailsMixin
from .parsers.output import ParserOutputMixin
from .parsers.outputs import ParserOutputsMixin
from .parsers.utils import mapping_or_none, str_or_none
from .template_precompile import DEFAULT_RENDERED_YAML_MAX_LEN, maybe_precompile_yaml_text
from .yaml_load import envelope_from_validation_issue, error_loc_for_yaml_path, load_yaml_mapping_text

__all__ = []


def _create_validator() -> "ConfigValidator":
    from .validator import ConfigValidator  # noqa: PLC0415

    return ConfigValidator()


class YamlDemandLoader(
    ParserFieldsMixin,
    ParserOutputsMixin,
    ParserOutputMixin,
    ParserGuardrailsMixin,
):
    _validator: Optional["ConfigValidator"]

    def __init__(self) -> None:
        self._validator = None

    def load(
        self,
        source: Union[str, Path, IO[str]],
        *,
        template_vars: Optional[Mapping[str, object]] = None,
        template_sandbox: str = "safe",
        rendered_yaml_max_len: int = DEFAULT_RENDERED_YAML_MAX_LEN,
        allowed_yaml_roots: Optional[Sequence[Union[str, Path]]] = None,
        scalim_yaml_override: Optional[Union[str, Path]] = None,
        project_root_override: Optional[Union[str, Path]] = None,
    ) -> DemandConfig:
        yaml_path: Optional[Path] = None
        if isinstance(source, (str, Path)):
            yaml_path = Path(source)
            text = yaml_path.read_text(encoding=UTF8_ENCODING)
            text = maybe_precompile_yaml_text(
                text,
                template_vars=template_vars,
                context_label="需求 `YAML` 文件 `{}`".format(str(yaml_path)),
                context_kind="demand",
                template_sandbox=template_sandbox,
                rendered_yaml_max_len=rendered_yaml_max_len,
            )
            data, locations, _lines = load_yaml_mapping_text(
                text,
                source_path=str(yaml_path),
                detect_duplicate_keys=True,
            )
        else:
            text = source.read()
            text = maybe_precompile_yaml_text(
                text,
                template_vars=template_vars,
                context_label="需求 `YAML` 文本",
                context_kind="demand",
                template_sandbox=template_sandbox,
                rendered_yaml_max_len=rendered_yaml_max_len,
            )
            data, locations, _lines = load_yaml_mapping_text(
                text,
                source_path="(inline)",
                detect_duplicate_keys=True,
            )
        raw_demand = RawDemand.from_raw(data)

        if contains_import_syntax(raw_demand.data):
            if yaml_path is not None:
                try:
                    _ = expand_imports_inplace(
                        raw_demand.data,
                        yaml_path=yaml_path,
                        template_vars=template_vars,
                        template_sandbox=template_sandbox,
                        rendered_yaml_max_len=rendered_yaml_max_len,
                        allowed_yaml_roots=allowed_yaml_roots,
                        scalim_yaml_override=scalim_yaml_override,
                        project_root_override=project_root_override,
                    )
                except ScalimYamlImportExpansionError as exc:
                    logical_path = str(exc.logical_path or "(root)")
                    error_message = "YAML imports expansion failed"
                    raise ScalimYamlValidationError(
                        error_message,
                        errors=[
                            ErrorEnvelope(
                                code="yaml_import_expansion_error",
                                message=str(exc),
                                source_path=str(yaml_path),
                                path=logical_path,
                                loc=error_loc_for_yaml_path(logical_path, locations),
                            )
                        ],
                    ) from None
            else:
                msg = "imports/$import is only supported for file path entrypoints; use YamlDemandLoader.load(<yaml_path>)"
                error_message = "YAML imports expansion is not supported for inline YAML"
                raise ScalimYamlValidationError(
                    error_message,
                    errors=[
                        ErrorEnvelope(
                            code="yaml_import_unsupported",
                            message=msg,
                            source_path="(inline)",
                            path="(root)",
                            loc=error_loc_for_yaml_path("(root)", locations),
                        )
                    ],
                )

        self._ensure_validator()
        if self._validator:
            report = self._validator.validate_report(
                raw_demand.data,
                strict_unknown_fields=True,
                enable_jsonschema_validation=True,
            )
            errors = [
                envelope_from_validation_issue(
                    issue,
                    source_path=str(yaml_path) if yaml_path is not None else "(inline)",
                    locations=locations,
                    default_code="yaml_validate_error",
                )
                for issue in report.errors()
            ]
            warnings = [
                envelope_from_validation_issue(
                    issue,
                    source_path=str(yaml_path) if yaml_path is not None else "(inline)",
                    locations=locations,
                    default_code="yaml_validate_warning",
                )
                for issue in report.warnings()
            ]
            if errors:
                error_message = "YAML DSL validation failed"
                raise ScalimYamlValidationError(
                    error_message,
                    errors=errors,
                    warnings=warnings,
                )

        return self._parse_config(raw_demand)

    def load_string(
        self,
        yaml_string: str,
        *,
        template_vars: Optional[Mapping[str, object]] = None,
        template_sandbox: str = "safe",
        rendered_yaml_max_len: int = DEFAULT_RENDERED_YAML_MAX_LEN,
    ) -> DemandConfig:
        text = maybe_precompile_yaml_text(
            yaml_string,
            template_vars=template_vars,
            context_label="需求 `YAML` 字符串",
            context_kind="demand",
            template_sandbox=template_sandbox,
            rendered_yaml_max_len=rendered_yaml_max_len,
        )
        data, locations, _lines = load_yaml_mapping_text(
            text,
            source_path="(inline)",
            detect_duplicate_keys=True,
        )
        raw_demand = RawDemand.from_raw(data)

        if contains_import_syntax(raw_demand.data):
            msg = "imports/$import is only supported for file path entrypoints; use YamlDemandLoader.load(<yaml_path>)"
            error_message = "YAML imports expansion is not supported for inline YAML"
            raise ScalimYamlValidationError(
                error_message,
                errors=[
                    ErrorEnvelope(
                        code="yaml_import_unsupported",
                        message=msg,
                        source_path="(inline)",
                        path="(root)",
                        loc=error_loc_for_yaml_path("(root)", locations),
                    )
                ],
            )

        self._ensure_validator()
        if self._validator:
            report = self._validator.validate_report(
                raw_demand.data,
                strict_unknown_fields=True,
                enable_jsonschema_validation=True,
            )
            errors = [
                envelope_from_validation_issue(
                    issue,
                    source_path="(inline)",
                    locations=locations,
                    default_code="yaml_validate_error",
                )
                for issue in report.errors()
            ]
            warnings = [
                envelope_from_validation_issue(
                    issue,
                    source_path="(inline)",
                    locations=locations,
                    default_code="yaml_validate_warning",
                )
                for issue in report.warnings()
            ]
            if errors:
                error_message = "YAML DSL validation failed"
                raise ScalimYamlValidationError(
                    error_message,
                    errors=errors,
                    warnings=warnings,
                )

        return self._parse_config(raw_demand)

    def _ensure_validator(self) -> None:
        if self._validator is None:
            self._validator = _create_validator()

    def _parse_config(self, raw: RawDemand) -> DemandConfig:
        name = str(raw.data.get(DEMAND_KEYS["name"], ""))
        description = str(raw.data.get(DEMAND_KEYS["description"], ""))
        if DEMAND_KEYS["batch_size"] in raw.data:
            batch_size = raw.data.get(DEMAND_KEYS["batch_size"])
        else:
            batch_size = DEFAULT_BATCH_SIZE
        retry = self._parse_loader_retry(raw.data.get(DEMAND_KEYS["retry"]))
        main_source = self._parse_main_source(raw)
        sources = self._parse_sources(raw)
        relations = self._parse_relations(raw)
        field_def_index = self._collect_field_defs(raw, main_source.source_id)
        resources = self._parse_resources(raw)
        outputs, required_field_ids = self._parse_outputs(raw, field_def_index=field_def_index)

        parsed_fields = self._parse_fields(
            raw,
            main_source.source_id,
            required_field_ids,
            relations,
            field_def_index=field_def_index,
        )
        main_source = self._with_main_source_fields(main_source, parsed_fields.main_source_fields)
        sources = self._with_source_fields(sources, parsed_fields.source_fields_by_source)

        validate_unique_raw = raw.data.get(DEMAND_KEYS["validate_unique_field_names"])
        if validate_unique_raw is None:
            validate_unique_field_names = True
        elif isinstance(validate_unique_raw, bool):
            validate_unique_field_names = bool(validate_unique_raw)
        else:
            msg = "validate_unique_field_names must be a boolean"
            raise TypeError(msg)

        failure_policy = str(raw.data.get(DEMAND_KEYS["failure_policy"], "all_fail") or "all_fail")
        if failure_policy not in ("all_fail", "primary_only"):
            msg = "failure_policy must be 'all_fail' or 'primary_only'"
            raise ValueError(msg)

        include_full_error_message = bool(raw.data.get(DEMAND_KEYS["include_full_error_message"], False))
        meta = self._parse_extra_sheet(raw.data.get(DEMAND_KEYS["meta"]), key="meta")
        audit = self._parse_extra_sheet(raw.data.get(DEMAND_KEYS["audit"]), key="audit")

        observability = self._parse_observability(raw.data)
        guardrails = self._parse_guardrails(raw.data, parsed_fields.field_def_index)

        return DemandConfig(
            name=name,
            description=description,
            batch_size=batch_size,
            retry=retry,
            main_source=main_source,
            sources=sources,
            source_fields=parsed_fields.source_fields,
            derived_fields=parsed_fields.derived_fields,
            source_field_id_map=parsed_fields.source_field_id_map,
            relations=relations,
            guardrails=guardrails,
            resources=resources,
            outputs=outputs,
            validate_unique_field_names=validate_unique_field_names,
            failure_policy=failure_policy,
            include_full_error_message=include_full_error_message,
            meta=meta,
            audit=audit,
            observability=observability,
        )

    def _parse_resources(self, raw: RawDemand) -> Optional[ResourcesConfig]:  # noqa: C901, PLR0912
        resources_dict = raw.get_mapping(DEMAND_KEYS["resources"])
        if resources_dict is None:
            return None

        books: Dict[str, BookConfig] = {}
        books_dict = mapping_or_none(resources_dict.get(RESOURCES_KEYS["books"]))
        if books_dict is None:
            if RESOURCES_KEYS["books"] in resources_dict:
                msg = "resources.books must be an object"
                raise TypeError(msg)
        else:
            for raw_book_id, raw_book_cfg in books_dict.items():
                book_id = str(raw_book_id or "").strip()
                if not book_id:
                    msg = "resources.books key must be a non-empty string"
                    raise ValueError(msg)

                book_cfg_dict = mapping_or_none(raw_book_cfg)
                if book_cfg_dict is None:
                    msg = "resources.books.{} must be an object".format(book_id)
                    raise TypeError(msg)

                books[book_id] = self._parse_book_config(book_cfg_dict, base_path="resources.books.{}".format(book_id))

        files: Dict[str, FileConfig] = {}
        files_dict = mapping_or_none(resources_dict.get(RESOURCES_KEYS["files"]))
        if files_dict is None:
            if RESOURCES_KEYS["files"] in resources_dict:
                msg = "resources.files must be an object"
                raise TypeError(msg)
        else:
            for raw_file_id, raw_file_cfg in files_dict.items():
                file_id = str(raw_file_id or "").strip()
                if not file_id:
                    msg = "resources.files key must be a non-empty string"
                    raise ValueError(msg)

                file_cfg_dict = mapping_or_none(raw_file_cfg)
                if file_cfg_dict is None:
                    msg = "resources.files.{} must be an object".format(file_id)
                    raise TypeError(msg)

                files[file_id] = self._parse_file_config(file_cfg_dict, base_path="resources.files.{}".format(file_id))

        return ResourcesConfig(books=books, files=files)

    def _parse_book_config(self, raw: Dict[str, Any], *, base_path: str) -> BookConfig:  # noqa: C901
        kind = str(raw.get(BOOK_KEYS["kind"]) or "").strip()
        if not kind:
            msg = "{}.kind is required".format(base_path)
            raise ValueError(msg)
        if kind not in BOOK_KINDS:
            msg = "{}.kind={!r} is invalid; expected one of: {}".format(base_path, kind, ", ".join(BOOK_KINDS))
            raise ValueError(msg)

        path = self._parse_path_or_init_var(raw.get(BOOK_KEYS["path"]), path="{}.path".format(base_path))

        budget_cfg = None
        budget_raw = mapping_or_none(raw.get(BOOK_KEYS["budget"]))
        if budget_raw is not None:
            budget_cfg = self._parse_book_budget(budget_raw, base_path="{}.budget".format(base_path))

        export_cfg = None
        export_raw = mapping_or_none(raw.get(BOOK_KEYS["export_xlsx"]))
        if export_raw is not None:
            export_cfg = self._parse_book_export_xlsx(export_raw, base_path="{}.export_xlsx".format(base_path))

        allow_formulas = bool(raw.get(BOOK_KEYS["allow_formulas"], False))
        write_lock = bool(raw.get(BOOK_KEYS["write_lock"], False))

        write_defaults_cfg = None
        write_defaults_raw = mapping_or_none(raw.get(BOOK_KEYS["write_defaults"]))
        if write_defaults_raw is not None:
            write_defaults_cfg = self._parse_book_write_defaults(write_defaults_raw, base_path="{}.write_defaults".format(base_path))

        # 语义层防御性校验(即使 `schema` 已覆盖,仍保持 `fail-fast` 便于诊断).
        if kind == "xlsx_file":
            if not path or (isinstance(path, str) and not path.strip()):
                msg = "{}.path is required for kind=xlsx_file".format(base_path)
                raise ValueError(msg)
            if budget_cfg is not None:
                msg = "{}.budget is not allowed for kind=xlsx_file".format(base_path)
                raise ValueError(msg)
            if export_cfg is not None:
                msg = "{}.export_xlsx is not allowed for kind=xlsx_file".format(base_path)
                raise ValueError(msg)
        if kind == "xlsx_memory":
            if budget_cfg is None:
                msg = "{}.budget is required for kind=xlsx_memory".format(base_path)
                raise ValueError(msg)
            if path is not None:
                msg = "{}.path is not allowed for kind=xlsx_memory".format(base_path)
                raise ValueError(msg)

        return BookConfig(
            kind=kind,
            path=path,
            budget=budget_cfg,
            export_xlsx=export_cfg,
            allow_formulas=allow_formulas,
            write_lock=write_lock,
            write_defaults=write_defaults_cfg,
        )

    def _parse_file_config(self, raw: Dict[str, Any], *, base_path: str) -> FileConfig:
        kind = str(raw.get(FILE_KEYS["kind"]) or "").strip()
        if not kind:
            msg = "{}.kind is required".format(base_path)
            raise ValueError(msg)
        if kind not in FILE_KINDS:
            msg = "{}.kind={!r} is invalid; expected one of: {}".format(base_path, kind, ", ".join(FILE_KINDS))
            raise ValueError(msg)

        path = self._parse_path_or_init_var(raw.get(FILE_KEYS["path"]), path="{}.path".format(base_path))
        if not path or (isinstance(path, str) and not path.strip()):
            msg = "{}.path is required for kind=csv_file".format(base_path)
            raise ValueError(msg)

        encoding = str(raw.get(FILE_KEYS["encoding"]) or "").strip() or UTF8_ENCODING
        return FileConfig(kind=kind, path=path, encoding=encoding)

    def _parse_path_or_init_var(self, raw: object, *, path: str) -> Any:
        if isinstance(raw, dict):
            return {
                "$init_var": parse_init_var_mapping_node(
                    cast("Dict[str, Any]", raw),  # pragma: allow-cast init_var mapping typed narrowing
                    path=path,
                )
            }
        if raw is None:
            return None
        if isinstance(raw, str):
            return raw.strip()
        msg = "{} must be a non-empty string or {{$init_var: <name>}}".format(path)
        raise TypeError(msg)

    def _parse_book_budget(self, raw: Dict[str, Any], *, base_path: str) -> BookBudgetConfig:
        max_sheets_raw = raw.get(BOOK_BUDGET_KEYS["max_sheets"])
        max_total_cells_raw = raw.get(BOOK_BUDGET_KEYS["max_total_cells"])
        if max_sheets_raw is None:
            msg = "{}.max_sheets must be an integer".format(base_path)
            raise TypeError(msg)
        if max_total_cells_raw is None:
            msg = "{}.max_total_cells must be an integer".format(base_path)
            raise TypeError(msg)
        try:
            max_sheets = int(max_sheets_raw)
        except (TypeError, ValueError):
            msg = "{}.max_sheets must be an integer".format(base_path)
            raise TypeError(msg) from None
        try:
            max_total_cells = int(max_total_cells_raw)
        except (TypeError, ValueError):
            msg = "{}.max_total_cells must be an integer".format(base_path)
            raise TypeError(msg) from None
        if max_sheets < 1:
            msg = "{}.max_sheets must be >= 1".format(base_path)
            raise ValueError(msg)
        if max_total_cells < 1:
            msg = "{}.max_total_cells must be >= 1".format(base_path)
            raise ValueError(msg)
        return BookBudgetConfig(max_sheets=max_sheets, max_total_cells=max_total_cells)

    def _parse_book_export_xlsx(self, raw: Dict[str, Any], *, base_path: str) -> BookExportXlsxConfig:
        export_path = self._parse_path_or_init_var(raw.get(BOOK_EXPORT_XLSX_KEYS["path"]), path="{}.path".format(base_path))
        if not export_path or (isinstance(export_path, str) and not export_path.strip()):
            msg = "{}.path is required".format(base_path)
            raise ValueError(msg)
        write_lock = bool(raw.get(BOOK_EXPORT_XLSX_KEYS["write_lock"], False))
        allow_formulas = bool(raw.get(BOOK_EXPORT_XLSX_KEYS["allow_formulas"], False))
        return BookExportXlsxConfig(path=export_path, write_lock=write_lock, allow_formulas=allow_formulas)

    def _parse_book_write_defaults(self, raw: Dict[str, Any], *, base_path: str) -> BookWriteDefaultsConfig:
        mode = str(raw.get(BOOK_WRITE_DEFAULTS_KEYS["mode"]) or DEFAULT_BOOK_WRITE_MODE).strip() or DEFAULT_BOOK_WRITE_MODE
        if mode not in BOOK_WRITE_MODE_ENUM:
            msg = "{}.mode={!r} is invalid; expected one of: {}".format(base_path, mode, ", ".join(BOOK_WRITE_MODE_ENUM))
            raise ValueError(msg)

        align_by = str(raw.get(BOOK_WRITE_DEFAULTS_KEYS["align_by"]) or DEFAULT_BOOK_WRITE_ALIGN_BY).strip() or DEFAULT_BOOK_WRITE_ALIGN_BY
        if align_by not in BOOK_WRITE_ALIGN_BY_ENUM:
            msg = "{}.align_by={!r} is invalid; expected one of: {}".format(base_path, align_by, ", ".join(BOOK_WRITE_ALIGN_BY_ENUM))
            raise ValueError(msg)

        header_policy = (
            str(raw.get(BOOK_WRITE_DEFAULTS_KEYS["header_policy"]) or DEFAULT_BOOK_WRITE_HEADER_POLICY).strip()
            or DEFAULT_BOOK_WRITE_HEADER_POLICY
        )
        if header_policy not in BOOK_WRITE_HEADER_POLICY_ENUM:
            msg = "{}.header_policy={!r} is invalid; expected one of: {}".format(
                base_path, header_policy, ", ".join(BOOK_WRITE_HEADER_POLICY_ENUM)
            )
            raise ValueError(msg)

        on_mismatch = (
            str(raw.get(BOOK_WRITE_DEFAULTS_KEYS["on_mismatch"]) or DEFAULT_BOOK_WRITE_ON_MISMATCH).strip()
            or DEFAULT_BOOK_WRITE_ON_MISMATCH
        )
        if on_mismatch not in BOOK_WRITE_ON_MISMATCH_ENUM:
            msg = "{}.on_mismatch={!r} is invalid; expected one of: {}".format(
                base_path, on_mismatch, ", ".join(BOOK_WRITE_ON_MISMATCH_ENUM)
            )
            raise ValueError(msg)

        on_conflict = (
            str(raw.get(BOOK_WRITE_DEFAULTS_KEYS["on_conflict"]) or DEFAULT_BOOK_WRITE_ON_CONFLICT).strip()
            or DEFAULT_BOOK_WRITE_ON_CONFLICT
        )
        if on_conflict not in BOOK_WRITE_ON_CONFLICT_ENUM:
            msg = "{}.on_conflict={!r} is invalid; expected one of: {}".format(
                base_path, on_conflict, ", ".join(BOOK_WRITE_ON_CONFLICT_ENUM)
            )
            raise ValueError(msg)

        return BookWriteDefaultsConfig(
            mode=mode,
            align_by=align_by,
            header_policy=header_policy,
            on_mismatch=on_mismatch,
            on_conflict=on_conflict,
        )

    def _parse_extra_sheet(self, raw_value: object, *, key: str) -> Optional[OutputExtraSheetConfig]:
        if raw_value is None or raw_value is False:
            return None
        if raw_value is True:
            return OutputExtraSheetConfig()

        sheet_dict = mapping_or_none(raw_value)
        if sheet_dict is None:
            msg = "{} must be a boolean or an object".format(key)
            raise TypeError(msg)

        return OutputExtraSheetConfig(
            path=str_or_none(sheet_dict.get(OUTPUT_EXTRA_SHEET_KEYS["path"])),
            sheet=str_or_none(sheet_dict.get(OUTPUT_EXTRA_SHEET_KEYS["sheet"])),
            allow_formulas=sheet_dict.get(OUTPUT_EXTRA_SHEET_KEYS["allow_formulas"]),
            write_lock=sheet_dict.get(OUTPUT_EXTRA_SHEET_KEYS["write_lock"]),
        )



# region imports

from typing import TYPE_CHECKING, Dict, Hashable, List, Mapping, Optional, Sequence, Type, Union

from ...typedefs import FieldValue, RowData, SinkRowKeySeq
from ...vendor.compact.typing_extensionsx import Self, override
from .base import BaseRowSink, IColumnSink

if TYPE_CHECKING:
    import types

# endregion


class InMemoryRowSink(BaseRowSink):
    """内存行写入器:支持行级写入并存储在内存中(FR023).

    主要用于测试与调试.

    示例:
    ```python
    sink = InMemoryRowSink()
    sink.write_row({"id": 1, "name": "张三"})
    sink.write_batch([{"id": 2, "name": "李四"}])
    data = sink.get_data()
    ```
    """

    def __init__(self) -> None:
        self._data: List[RowData] = []
        self._closed: bool = False

    @override
    def write_row(self, row: RowData) -> None:
        self._data.append(dict(row))

    def write_row_aligned(self, field_keys: Sequence[str], values: Sequence[FieldValue]) -> None:
        if len(field_keys) != len(values):
            msg = "`write_row_aligned` 长度不一致: field_keys={} values={}".format(len(field_keys), len(values))
            raise ValueError(msg)
        self._data.append(dict(zip(field_keys, values)))

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        for row in rows:
            self._data.append(dict(row))

    @override
    def close(self) -> None:
        self._closed = True

    def get_data(self) -> List[RowData]:
        return self._data


InMemoryListSink = InMemoryRowSink


class InMemoryColumnSink(IColumnSink):
    """内存列写入器:支持按列追加写入并存储在内存中(FR023).

    主要用于测试、调试,以及需要在内存中进一步处理数据的场景.

    数据访问方式:
    - `get_columns()`:获取列式数据(`Dict[field_key, Dict[pk, value]]`)
    - `get_rows()`:获取行式数据(`List[Dict[field_key, value]]`)
    - `get_2d_list()`:获取二维列表(`List[List[value]]`)
    - `get_column(field_key)`:获取单列数据

    示例:
    ```python
    sink = InMemoryColumnSink(["order_id", "name"])
    sink.set_row_ids([1, 2, 3])
    sink.write_column("order_id", {1: 1, 2: 2, 3: 3})
    sink.write_column("name", {1: "甲", 2: "乙", 3: "丙"})
    sink.close()

    columns = sink.get_columns()
    rows = sink.get_rows()
    ```
    """

    field_names: List[str]
    _row_ids: List[Hashable]
    _columns: Dict[str, Dict[Hashable, FieldValue]]
    _closed: bool
    _auto_field_names: bool

    def __init__(self, field_names: Optional[List[str]] = None) -> None:
        self._auto_field_names = field_names is None
        self.field_names = field_names if field_names is not None else []
        self._row_ids = []
        self._columns = {}
        self._closed = False

    @override
    def set_row_ids(self, row_ids: "SinkRowKeySeq") -> None:
        self._row_ids.extend(row_ids)

    @override
    def write_column(self, field_key: str, values: Mapping[Hashable, FieldValue]) -> None:
        if field_key not in self._columns:
            self._columns[field_key] = {}
        self._columns[field_key].update(values)
        if self._auto_field_names and field_key not in self.field_names:
            self.field_names.append(field_key)

    def write_column_aligned(self, field_key: str, row_ids: "SinkRowKeySeq", values: Sequence[FieldValue]) -> None:
        if len(row_ids) != len(values):
            msg = "`write_column_aligned` 长度不一致: row_ids={} values={}".format(len(row_ids), len(values))
            raise ValueError(msg)

        if field_key not in self._columns:
            self._columns[field_key] = {}
        col = self._columns[field_key]
        for row_id, value in zip(row_ids, values):
            col[row_id] = value

        if self._auto_field_names and field_key not in self.field_names:
            self.field_names.append(field_key)

    @override
    def write_columns(self, columns: Mapping[str, Mapping[Hashable, FieldValue]]) -> None:
        for field_key, values in columns.items():
            self.write_column(field_key, values)

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        for row_idx, row in enumerate(rows):
            pk = row_idx
            if pk not in self._row_ids:
                self._row_ids.append(pk)
            for field_key, value in row.items():
                if field_key not in self._columns:
                    self._columns[field_key] = {}
                self._columns[field_key][pk] = value

    @override
    def close(self) -> None:
        self._closed = True

    # ============================================================
    # 数据访问方法
    # ============================================================

    def get_columns(self) -> Dict[str, Dict[Hashable, FieldValue]]:
        return self._columns

    def get_column(self, field_key: str) -> Dict[Hashable, FieldValue]:
        return self._columns.get(field_key, {})

    def get_rows(self) -> List[RowData]:
        rows: List[RowData] = []
        for pk in self._row_ids:
            row: Dict[str, FieldValue] = {}
            for field_key in self._columns:
                if pk in self._columns[field_key]:
                    row[field_key] = self._columns[field_key][pk]
            rows.append(row)
        return rows

    def get_2d_list(
        self,
        *,
        include_header: bool = False,
    ) -> List[List[Union[str, FieldValue]]]:
        result: List[List[Union[str, FieldValue]]] = []
        fields = self.field_names or list(self._columns.keys())

        if include_header:
            result.append(list(fields))

        for pk in self._row_ids:
            row_values: List[Union[str, FieldValue]] = []
            for field_key in fields:
                column_data = self._columns.get(field_key, {})
                row_values.append(column_data.get(pk))
            result.append(row_values)

        return result

    def get_row_ids(self) -> List[Hashable]:
        return self._row_ids

    def get_field_names(self) -> List[str]:
        return self.field_names or list(self._columns.keys())

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional["types.TracebackType"],  # noqa: PYI036
    ) -> None:
        self.close()


__all__ = []



from typing import List, Optional, Tuple, cast

from ...typedefs import LoaderResultMapping, RuntimeValue
from ._relations import JoinConditionIr, LookupStepIr, RelationIr
from ._sources import SourceIr, SourceRefIr
from .aliases import LoaderResultMapCallable
from .binding import BindingIr, LoaderCallContextIr


def infer_lookup_steps(
    relation: object,
    from_source: SourceRefIr,
    to_source: SourceIr,
) -> Optional[Tuple[LookupStepIr, ...]]:
    """
    从 `relation` 推断 `LookupStepIr` 序列

    这是一个统一的辅助函数,处理 `JoinConditionIr` 和 `RelationIr` 到 `LookupStepIr` 的转换

    参数:
        `relation`: 关联关系 (`JoinConditionIr` 或 `RelationIr`)
        `from_source`: 起始数据源
        `to_source`: 目标数据源

    注意:
        关联依赖方向与主数据源选择有关.推荐通过 `PlanBuilder`/`ExecutionPlan` 使用该推断逻辑,
        避免在 `IR` 层手工拼装导致方向不一致.

    返回:
        `LookupStepIr` 序列,无法推断时返回 `None`

    示例:
        - 单字段关联:
          `relation = orders[\"customer_id\"].join(customers[\"customer_id\"])`
          `steps = infer_lookup_steps(relation, orders_source, customers_source)`
        - 多级关联:
          `relation = orders[\"pay_id\"].join(pays[\"pay_id\"]).and_(pays[\"country_id\"].join(countries[\"country_id\"]))`
          `steps = infer_lookup_steps(relation, orders_source, countries_source)`
    """

    try:
        if isinstance(relation, JoinConditionIr):
            # 单个条件,转换为 `RelationIr`
            rel = RelationIr(conditions=(relation,))
            return rel.infer_multi_field_lookup_path(from_source, to_source)
        if isinstance(relation, RelationIr):
            return relation.infer_multi_field_lookup_path(from_source, to_source)
    except ValueError:
        return None

    return None


def extract_from_fields(
    steps: Tuple[LookupStepIr, ...],
) -> Tuple[str, ...]:
    """
    从 `LookupStepIr` 序列中收集所有步骤的 `from_fields`,用于依赖分析

    参数:
        `steps`: `LookupStepIr` 序列

    返回:
        所有 `from_fields` 的扁平化元组

    示例:
        - 单级单字段:
          `steps = (LookupStepIr(from_field=\"customer_id\", to_source=...),)`
          `extract_from_fields(steps)`  # -> `(\"customer_id\",)`
        - 多级:
          `steps = (LookupStepIr(from_field=\"pay_id\", to_source=pays), LookupStepIr(from_field=\"country_id\", to_source=countries))`
          `extract_from_fields(steps)`  # -> `(\"pay_id\", \"country_id\")`
        - 多字段:
          `steps = (LookupStepIr(from_field=(\"region_id\", \"institution_id\"), to_source=...),)`
          `extract_from_fields(steps)`  # -> `(\"region_id\", \"institution_id\")`
    """
    result: List[str] = []
    for step in steps:
        from_fields = step.get_from_fields()  # 返回 `Tuple[str, ...]`
        result.extend(from_fields)
    return tuple(result)


def call_loader_with_binding(
    binding: Optional[BindingIr],
    context: LoaderCallContextIr,
    loader_fn: LoaderResultMapCallable,
) -> RuntimeValue:
    if binding is None:
        return loader_fn()

    args, kwargs = binding.build_params(context)
    return loader_fn(*args, **kwargs)


def coerce_loader_result_mapping(result: object) -> LoaderResultMapping:
    return cast("LoaderResultMapping", result)  # pragma: allow-cast loader result boundary typed narrowing


__all__ = []



# region imports

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional, Set

from ..events import (
    EVENT_ADAPTIVE_SCHEDULER_DECISION,
    EVENT_BATCH_END,
    EVENT_BATCH_START,
    EVENT_COLUMN_WRITE,
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_ERROR,
    EVENT_FIELD_COMPUTE,
    EVENT_FIELD_SLIM,
    EVENT_LOADER_CALL,
    EVENT_LOADER_RETRY,
    EVENT_LOADER_SLIM,
    EVENT_OUTPUT_TARGET_END,
    EVENT_PIPELINE_END,
    EVENT_PIPELINE_START,
    EVENT_RELATION_LOOKUP,
    EVENT_ROW_RELEASE,
    EVENT_ROW_WRITE,
    EVENT_STAGE_SPAN,
    Event,
)
from ..vendor.compact.typing_extensionsx import override

# endregion

_DISPATCH_MAP = {
    EVENT_PIPELINE_START: "on_pipeline_start",
    EVENT_PIPELINE_END: "on_pipeline_end",
    EVENT_BATCH_START: "on_batch_start",
    EVENT_BATCH_END: "on_batch_end",
    EVENT_LOADER_CALL: "on_loader_call",
    EVENT_LOADER_RETRY: "on_loader_retry",
    EVENT_FIELD_COMPUTE: "on_field_compute",
    EVENT_ERROR: "on_error",
    EVENT_DIAGNOSTIC_WARNING: "on_diagnostic_warning",
    EVENT_FIELD_SLIM: "on_field_slim",
    EVENT_ROW_WRITE: "on_row_write",
    EVENT_ROW_RELEASE: "on_row_release",
    EVENT_LOADER_SLIM: "on_loader_slim",
    EVENT_COLUMN_WRITE: "on_column_write",
    EVENT_RELATION_LOOKUP: "on_relation_lookup",
    EVENT_STAGE_SPAN: "on_stage_span",
    EVENT_ADAPTIVE_SCHEDULER_DECISION: "on_adaptive_scheduler_decision",
    EVENT_OUTPUT_TARGET_END: "on_output_target_end",
}


class Observer(ABC):
    """观测器插件基类."""

    event_types: Optional[Set[str]] = None
    supports_unknown_event_types: bool = False

    def supports(self, event_type: str) -> bool:
        if self.event_types is None:
            return True
        return event_type in self.event_types

    @abstractmethod
    def on_event(self, event: Event) -> None:
        """处理统一事件."""

    def close(self) -> None:  # noqa: B027
        """可选的清理钩子."""


class EventDispatchObserver(Observer):
    """将事件分发到已定义的类型化处理函数."""

    dispatch_map: Dict[str, str] = _DISPATCH_MAP
    _handler_cache: Dict[str, Optional[Callable[[Any], Any]]]

    @override
    def on_event(self, event: Event) -> None:
        try:
            handler_cache = self._handler_cache
        except AttributeError:
            handler_cache = {}
            self._handler_cache = handler_cache

        event_type = event.event_type
        if event_type not in handler_cache:
            handler_name = self.dispatch_map.get(event_type)
            if not handler_name:
                handler_cache[event_type] = None
                return
            resolved = getattr(self, handler_name, None)  # pragma: allow-dynattr dispatch: handler_name
            if resolved is None or not callable(resolved):
                handler_cache[event_type] = None
                return
            handler_cache[event_type] = resolved

        handler = handler_cache[event_type]
        if handler is None:
            return
        _ = handler(event.payload)


__all__ = [
    "EventDispatchObserver",
    "Observer",
]



from __future__ import absolute_import

from typing import Optional, Tuple

from ..vendor.dataclassesx import dataclass


@dataclass(frozen=True)
class ExportLayout:
    """与 `DSL` 无关的导出布局.

    此对象定义:
    - 需要导出的字段(及其顺序)
    - 与字段顺序对齐的可选表头名称
    """

    field_ids: Tuple[str, ...]
    header_names: Optional[Tuple[str, ...]] = None

    def __post_init__(self) -> None:
        if self.header_names is not None and len(self.header_names) != len(self.field_ids):
            msg = "ExportLayout.header_names must align with field_ids"
            raise ValueError(msg)


@dataclass(frozen=True)
class OutputSpec:
    """与 `DSL` 无关的输出策略.

    当 `path` 为假值(`None`/`\"\"`)时,不会创建文件输出端.

    安全提示:
    - `path` 控制文件系统写入(会创建父目录并以原子方式替换目标文件).
    - 仅在 `YAML`/配置输入可信时启用基于文件的输出;否则应在外部校验/覆盖 `path`.
    """

    format: str = "csv"
    path: Optional[str] = None
    encoding: str = "utf-8"
    streaming: bool = True
    include_header: bool = True
    sheet_name: Optional[str] = None
    excel_allow_formulas: bool = False
    write_lock: bool = False


__all__ = [
    "ExportLayout",
    "OutputSpec",
]



from typing import Any, Dict, Hashable, Optional, Sequence, Tuple

from ..guardrails import GuardrailMode, ScalimGuardrailViolationError


def build_guardrail_once_key(code: str, *parts: Any) -> Tuple[str, ...]:
    return (code, *tuple(str(part) for part in parts))


def build_loader_row_guardrail_payload(
    runtime: Any,
    *,
    source_id: str,
    row_id: Hashable,
    field_key: Optional[str] = None,
    reason: Optional[str] = None,
    data_key: Optional[str] = None,
    error_type: Optional[str] = None,
    error: Optional[str] = None,
    is_ref_loader: bool = False,
    lookup_key: Any = None,
    include_lookup_key: bool = False,
    main_source: bool = False,
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {
        "source_id": source_id,
        "batch_num": runtime.batch_num,
        "row_id": row_id,
    }

    if field_key is not None:
        payload["field_key"] = field_key
    if reason is not None:
        payload["reason"] = reason
    if data_key is not None:
        payload["data_key"] = data_key
    if error_type is not None:
        payload["error_type"] = error_type
    if error is not None:
        payload["error"] = error
    if include_lookup_key:
        payload["lookup_key"] = lookup_key
    if is_ref_loader:
        payload["is_ref_loader"] = True
    if main_source:
        payload["main_source"] = True

    return payload


def build_loader_result_guardrail_payload(
    runtime: Any,
    *,
    source_id: str,
    result: Any,
    is_ref_loader: bool = False,
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {
        "source_id": source_id,
        "batch_num": runtime.batch_num,
        "result_type": type(result).__name__,
    }
    if is_ref_loader:
        payload["is_ref_loader"] = True
    return payload


def build_compute_error_guardrail_payload(
    runtime: Any,
    *,
    field_key: str,
    row_id: Hashable,
    dependencies: Sequence[str],
    exc: Exception,
    include_error: bool = False,
    unexpected: bool = False,
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {
        "batch_num": runtime.batch_num,
        "field_key": field_key,
        "row_id": row_id,
        "dependencies": list(dependencies),
        "error_type": type(exc).__name__,
    }
    if include_error:
        payload["error"] = str(exc)
    if unexpected:
        payload["unexpected"] = True
    return payload


def build_guardrail_context(
    runtime: Any,
    *,
    code: str,
    action_mode: GuardrailMode,
    context: Dict[str, Any],
) -> Dict[str, Any]:
    payload = dict(context)
    payload["guardrail"] = True
    payload["guardrail_code"] = code
    payload["guardrail_mode"] = action_mode
    payload["guardrail_configured_mode"] = runtime.guardrails.mode
    return payload


def record_guardrail(
    runtime: Any,
    *,
    code: str,
    message: str,
    action_mode: GuardrailMode,
    context: Dict[str, Any],
    once_key: Optional[Tuple[str, ...]] = None,
) -> None:
    payload = build_guardrail_context(runtime, code=code, action_mode=action_mode, context=context)
    if once_key is not None:
        if once_key in runtime.guardrail_logged:
            return
        runtime.guardrail_logged.add(once_key)
    runtime.instrumentation.emit_error(ScalimGuardrailViolationError(message, code=code, context=payload), payload)


def fail_guardrail(
    runtime: Any,
    *,
    code: str,
    message: str,
    context: Dict[str, Any],
    action_mode: GuardrailMode = "fast_fail",
    cause: Optional[Exception] = None,
) -> None:
    payload = build_guardrail_context(runtime, code=code, action_mode=action_mode, context=context)
    if cause is not None:
        payload["cause_type"] = type(cause).__name__
        payload["cause"] = str(cause)
    violation = ScalimGuardrailViolationError(message, code=code, context=payload)
    runtime.instrumentation.emit_error(violation, payload)
    raise violation


__all__ = []



from typing import Dict, FrozenSet, Mapping, Sequence, Set, Tuple

from ..spec.ir._workflow import WorkflowIr
from ..vendor.dataclassesx import dataclass


def _validate_workflow_node_ids(*, deps_by_node_id: Mapping[str, Sequence[str]]) -> Set[str]:
    node_ids: Set[str] = set()
    for raw_node_id in deps_by_node_id:
        node_id = str(raw_node_id).strip()
        if not node_id:
            msg = "workflow node_id must be a non-empty string"
            raise ValueError(msg)
        if node_id in node_ids:
            msg = "workflow node_id duplicated: {!r}".format(node_id)
            raise ValueError(msg)
        node_ids.add(node_id)
    return node_ids


def _normalize_workflow_deps(*, deps_by_node_id: Mapping[str, Sequence[str]]) -> Dict[str, Tuple[str, ...]]:
    normalized: Dict[str, Tuple[str, ...]] = {}
    for raw_node_id, deps in deps_by_node_id.items():
        consumer = str(raw_node_id).strip()
        if not consumer:
            msg = "workflow node_id must be a non-empty string"
            raise ValueError(msg)

        tmp: Set[str] = set()
        for dep_id in deps or ():
            dep = str(dep_id).strip()
            if not dep:
                msg = "workflow deps must contain only non-empty strings (node_id={!r})".format(consumer)
                raise ValueError(msg)
            tmp.add(dep)
        normalized[consumer] = tuple(sorted(tmp))
    return normalized


def _validate_workflow_deps_exist(*, deps_by_node_id: Mapping[str, Sequence[str]], node_ids: Set[str]) -> None:
    for raw_consumer, deps in deps_by_node_id.items():
        consumer = str(raw_consumer).strip()
        for raw_dep in deps or ():
            dep = str(raw_dep).strip()
            if dep and dep not in node_ids:
                msg = "workflow node {!r} depends_on unknown node {!r}".format(consumer, dep)
                raise ValueError(msg)


def _build_workflow_visibility_closure(*, deps_by_node_id: Mapping[str, Tuple[str, ...]], node_ids: Set[str]) -> Dict[str, FrozenSet[str]]:
    cache: Dict[str, FrozenSet[str]] = {}
    visiting: Set[str] = set()

    def _visible(consumer_node_id: str) -> FrozenSet[str]:
        cached = cache.get(consumer_node_id)
        if cached is not None:
            return cached

        if consumer_node_id in visiting:
            msg = "workflow depends_on cycle detected at node_id={!r}".format(consumer_node_id)
            raise ValueError(msg)

        visiting.add(consumer_node_id)
        out: Set[str] = set()
        for dep in deps_by_node_id.get(consumer_node_id, ()):
            out.add(dep)
            out.update(_visible(dep))
        visiting.remove(consumer_node_id)

        frozen = frozenset(out)
        cache[consumer_node_id] = frozen
        return frozen

    return {node_id: _visible(node_id) for node_id in sorted(node_ids)}


@dataclass(frozen=True)
class WorkflowVisibilityIndex:
    """`workflow` 节点可见性闭包(`SSOT`).

    可见性规则:
    - 对于 `consumer` 节点,其可见 `producer` 集合为 `depends_on` 的传递闭包(不包含自身).
    - 该对象仅做纯计算/纯数据,用于在上下文引用与产物读取等处复用.
    """

    visible_by_consumer_node_id: Dict[str, FrozenSet[str]]

    @classmethod
    def build(cls, *, deps_by_node_id: Mapping[str, Sequence[str]]) -> "WorkflowVisibilityIndex":
        node_ids = _validate_workflow_node_ids(deps_by_node_id=deps_by_node_id)
        _validate_workflow_deps_exist(deps_by_node_id=deps_by_node_id, node_ids=node_ids)
        normalized_deps = _normalize_workflow_deps(deps_by_node_id=deps_by_node_id)
        visible_by_consumer = _build_workflow_visibility_closure(deps_by_node_id=normalized_deps, node_ids=node_ids)
        return cls(visible_by_consumer_node_id=visible_by_consumer)

    @classmethod
    def from_workflow_ir(cls, workflow_ir: WorkflowIr) -> "WorkflowVisibilityIndex":
        deps_by_node_id = {str(node.node_id): tuple(str(dep_id) for dep_id in node.deps) for node in workflow_ir.nodes}
        return cls.build(deps_by_node_id=deps_by_node_id)

    def visible_producer_node_ids(self, consumer_node_id: str) -> FrozenSet[str]:
        return self.visible_by_consumer_node_id.get(str(consumer_node_id), frozenset())


__all__ = []



"""`workflow` 加载阶段实现(内部模块).

说明:
- 负责读取 `workflow` `YAML` 并解析为 `config`
- 运行时需兼容 `Python 3.6`
"""

from typing import Mapping, Optional, Tuple

from ._internal.config_parsing.template_precompile import DEFAULT_RENDERED_YAML_MAX_LEN
from .workflow import ScalimWorkflowConfigError, WorkflowConfig, load_workflow_config


def load_workflow_config_from_path(
    workflow_yaml_path: str,
    *,
    template_vars: Optional[Mapping[str, object]] = None,
    template_sandbox: str = "safe",
    rendered_yaml_max_len: int = DEFAULT_RENDERED_YAML_MAX_LEN,
) -> Tuple[str, WorkflowConfig]:
    workflow_path = str(workflow_yaml_path or "").strip()
    if not workflow_path:
        msg = "workflow_yaml_path is required"
        raise ScalimWorkflowConfigError(msg, path="(file)")
    wf = load_workflow_config(
        workflow_path,
        template_vars=template_vars,
        template_sandbox=template_sandbox,
        rendered_yaml_max_len=rendered_yaml_max_len,
    )
    return workflow_path, wf


__all__ = [
    "load_workflow_config_from_path",
]



from typing import Dict, FrozenSet, List, Mapping, Optional, Union

from ....spec.ir import DemandIr, DerivedFieldIr, FieldIr, MainSourceIr, SourceIr
from .._internal.config_parsing.security import SecureComputeEngine
from ..schema_dsl.models import DemandConfig
from ._internal.conversion_lookup import LookupCastRegistry, validate_source_id
from ._internal.conversion_relations import StepInfo
from ._internal.conversion_sources import ConfigToIRConversionSourceMixin
from .errors import ALLOWLIST_REQUIRED_MSG, ScalimAllowlistRequiredError, ScalimConversionError
from .references import PythonReferenceResolver, SecurePythonReferenceResolver


def _validate_source_id(source_id: str, context: str) -> None:
    validate_source_id(source_id, context)


# 保持 `_validate_source_id` 可导入(用于测试/门禁),但不纳入 `__all__`.
_non_public_exports = (_validate_source_id,)
del _non_public_exports


class ConfigToIRConverter(ConfigToIRConversionSourceMixin):
    _resolver: Optional[PythonReferenceResolver]
    _compute_engine: Optional[SecureComputeEngine]
    _lookup_casts: Optional[LookupCastRegistry]
    _sources_ir: Optional[Dict[str, SourceIr]]
    _main_source_ir: Optional[MainSourceIr]
    _relation_steps: Optional[Dict[str, List[StepInfo]]]
    _relation_adjacency: Optional[Dict[str, List[StepInfo]]]
    _source_field_id_map: Optional[Dict[str, Dict[str, str]]]
    _source_data_key_map: Optional[Dict[str, Dict[str, List[str]]]]
    _init_vars: Optional[Mapping[str, object]]

    @classmethod
    def from_allowlist(
        cls,
        *,
        allowed_modules: Optional[FrozenSet[str]] = None,
        allowed_functions: Optional[FrozenSet[str]] = None,
        compute_engine: Optional[SecureComputeEngine] = None,
    ) -> "ConfigToIRConverter":
        if not allowed_modules and not allowed_functions:
            raise ScalimAllowlistRequiredError(ALLOWLIST_REQUIRED_MSG)
        resolver = SecurePythonReferenceResolver(
            allowed_modules=allowed_modules,
            allowed_functions=allowed_functions,
        )
        return cls(resolver=resolver, compute_engine=compute_engine)

    def __init__(
        self,
        resolver: Optional[PythonReferenceResolver] = None,
        compute_engine: Optional[SecureComputeEngine] = None,
        init_vars: Optional[Mapping[str, object]] = None,
    ) -> None:
        resolved = resolver
        if resolved is None or not resolved.has_allowlist():
            raise ScalimAllowlistRequiredError(ALLOWLIST_REQUIRED_MSG)

        self._resolver = resolved
        self._compute_engine = compute_engine or SecureComputeEngine()
        self._init_vars = init_vars
        self._lookup_casts = LookupCastRegistry()
        self._sources_ir = {}
        self._main_source_ir = None
        self._relation_steps = {}
        self._relation_adjacency = {}
        self._source_field_id_map = {}
        self._source_data_key_map = {}

    def convert(self, config: DemandConfig) -> DemandIr:
        self._sources_ir = {}
        self._main_source_ir = None
        self._relation_steps = {}
        self._relation_adjacency = {}

        source_field_id_map = config.source_field_id_map or {}
        self._source_field_id_map = source_field_id_map
        self._source_data_key_map = self._build_source_data_key_map(source_field_id_map)

        main_source_ir = self._convert_main_source(config.main_source)
        self._main_source_ir = main_source_ir

        sources_ir = self._require_sources_ir()
        for source_id, source_config in config.sources.items():
            sources_ir[source_id] = self._convert_source(source_config)

        if main_source_ir.source_id in sources_ir:
            msg = "Main source '{}' conflicts with sources".format(main_source_ir.source_id)
            raise ScalimConversionError(msg)

        relation_steps = self._convert_relations(config)
        self._relation_steps = relation_steps
        self._relation_adjacency = self._build_relation_adjacency(relation_steps)

        fields_ir: List[Union[FieldIr, DerivedFieldIr]] = []

        for field_config in config.source_fields.values():
            fields_ir.append(self._convert_source_field(field_config, config))

        for derived_config in config.derived_fields.values():
            fields_ir.append(self._convert_derived_field(derived_config))

        return DemandIr.from_irs(
            sources=list(sources_ir.values()),
            fields=fields_ir,
            main_source=main_source_ir,
            batch_size_hint=config.batch_size,
            name=config.name,
        )


__all__ = ["ConfigToIRConverter", "LookupCastRegistry"]



from decimal import Decimal, InvalidOperation
from typing import Any, Callable, Dict, FrozenSet, List, Optional, Sequence, Tuple, cast

from ....execution.output_composition import (
    AggMetricSpec,
    AuditSheetSpec,
    DerivedGroupBySpec,
    DerivedOutputTargetSpec,
    MetaSheetSpec,
    OutputCompositionSpec,
    OutputRowPredicate,
    OutputTargetSpec,
    PostFieldSpec,
    RankFieldSpec,
)
from ....execution.output_contracts import ExportLayout, OutputSpec
from ....execution.run_ir import export_layout_from_demand_ir
from ....spec.ir import DemandIr
from ....typedefs import FieldValue, RowData
from ....vendor.dataclassesx import dataclass
from .._internal.config_parsing.call_by import CallByValue, ParsedCallBy, ScalimCallByParseError, parse_call_by
from .._internal.config_parsing.security import ScalimComputeExpressionError, ScalimSecurityError, SecureComputeEngine
from ..schema_dsl.constants import DEFAULT_OUTPUT_HEADER_BY, DEFAULT_OUTPUT_INCLUDE_HEADER
from ..schema_dsl.models import (
    BookConfig,
    BookWriteDefaultsConfig,
    DemandConfig,
    FileConfig,
    OutputAggregateConfig,
    OutputExtraSheetConfig,
    OutputTargetConfig,
)
from ..schema_dsl.output_enums import (
    AGG_METRIC_PRODUCER_KEYS as _AGG_FUNC_KEYS,
)
from ..schema_dsl.output_enums import (
    AGG_POST_PRODUCER_KEYS as _POST_FUNC_KEYS,
)
from ..schema_dsl.output_enums import (
    AGG_RANK_PRODUCER_KEYS as _RANK_FUNC_KEYS,
)
from ..schema_dsl.output_enums import (
    DEFAULT_BOOK_WRITE_HEADER_POLICY,
    DEFAULT_BOOK_WRITE_MODE,
)
from .output_path_resolve import resolve_yaml_relative_output_path
from .references import SecurePythonReferenceResolver

_EXCEL_SHEET_NAME_MAX_LEN = 31
_EXCEL_SHEET_NAME_INVALID_CHARS = frozenset(["\\", "/", "?", "*", "[", "]", ":"])


def _validate_excel_sheet_name(sheet: str, *, path: str) -> None:
    name = str(sheet or "").strip()
    if not name:
        msg = "Excel sheet name must be non-empty"
        err = "{} (path={})".format(msg, path)
        raise ValueError(err)
    if len(name) > _EXCEL_SHEET_NAME_MAX_LEN:
        msg = "Excel sheet name is too long (max_len={})".format(_EXCEL_SHEET_NAME_MAX_LEN)
        err = "{} (path={})".format(msg, path)
        raise ValueError(err)
    invalid = sorted(set(name).intersection(_EXCEL_SHEET_NAME_INVALID_CHARS))
    if invalid:
        msg = "Excel sheet name contains invalid characters: {}".format("".join(invalid))
        err = "{} (path={})".format(msg, path)
        raise ValueError(err)


def _ensure_field_value(value: object, *, field_id: str, producer: str) -> FieldValue:
    if value is None:
        return None
    if isinstance(value, (int, float, Decimal, str, bool)):
        return cast("FieldValue", value)  # pragma: allow-cast literal typed narrowing
    msg = "aggregate field {!r} produced unsupported value type {} from {}".format(field_id, type(value).__name__, producer)
    raise TypeError(msg)


def _decimal_from_text(text: str) -> Optional[Decimal]:
    if not text:
        return None
    try:
        return Decimal(text)
    except InvalidOperation:
        return None


def _to_decimal(value: object) -> Optional[Decimal]:
    if value is None:
        return None
    dec: Optional[Decimal] = None
    if isinstance(value, Decimal):
        dec = value
    elif isinstance(value, bool):
        dec = Decimal(1) if value else Decimal(0)
    elif isinstance(value, int):
        dec = Decimal(value)
    elif isinstance(value, float):
        dec = _decimal_from_text(str(value))
    elif isinstance(value, str):
        dec = _decimal_from_text(value.strip())
    if dec is None:
        return None
    if not dec.is_finite():
        return None
    return dec


@dataclass(frozen=True)
class _AggregateCallByContext:
    row_id: Optional[object]
    batch_num: int
    field_id: str
    deps: Tuple[str, ...]
    values: Dict[str, FieldValue]


def _eval_call_by_value(*, field_id: str, value: CallByValue, row: RowData, ctx: _AggregateCallByContext) -> object:
    kind = str(value.kind)
    if kind == "literal":
        return value.value
    if kind == "field":
        return row.get(str(value.value))
    if kind == "ctx":
        return ctx
    if kind == "ctx_attr":
        return getattr(ctx, str(value.value))  # pragma: allow-dynattr dsl: ctx_attr access
    msg = "Unknown call_by value kind: {} (field_id={!r})".format(
        kind, field_id
    )  # pragma: no cover  # pragma: allow-no-cover invariant: exhaustive CallByValue kind
    raise ValueError(msg)


def _compile_call_by_post_field(
    *,
    out_field_id: str,
    call_by: str,
    resolver: SecurePythonReferenceResolver,
) -> PostFieldSpec:
    try:
        parsed = parse_call_by(call_by)
    except ScalimCallByParseError as exc:
        msg = "aggregate.fields.{} has invalid call_by: {}".format(out_field_id, exc)
        raise ValueError(msg) from exc

    try:
        fn = resolver.resolve(parsed.reference)
    except Exception as exc:
        msg = "aggregate.fields.{} failed to resolve call_by reference '{}': {}".format(out_field_id, parsed.reference, exc)
        raise ValueError(msg) from exc

    deps = tuple(str(x) for x in (parsed.field_names or ()))

    def calculator(row: RowData, p: ParsedCallBy = parsed, f: Callable[..., Any] = fn) -> FieldValue:
        dep_values: Dict[str, FieldValue] = {name: row.get(name) for name in deps}
        ctx = _AggregateCallByContext(
            row_id=None,
            batch_num=0,
            field_id=str(out_field_id),
            deps=deps,
            values=dep_values,
        )

        args: List[object] = []
        for arg_value in p.args:
            args.append(_eval_call_by_value(field_id=str(out_field_id), value=arg_value, row=row, ctx=ctx))

        kwargs: Dict[str, object] = {}
        for key, kw_value in p.kwargs:
            kwargs[str(key)] = _eval_call_by_value(field_id=str(out_field_id), value=kw_value, row=row, ctx=ctx)

        result = f(*args, **kwargs)
        return _ensure_field_value(result, field_id=str(out_field_id), producer="call_by")

    return PostFieldSpec(
        out_field_id=str(out_field_id),
        kind="call_by",
        dependencies=deps,
        fingerprint=str(call_by),
        calculator=calculator,
    )


def _compile_score_by_rank_post_field(
    *,
    out_field_id: str,
    cfg: Dict[str, Any],
) -> PostFieldSpec:
    rank_field = str(cfg.get("rank_field") or "rank").strip() or "rank"
    base_dec = _to_decimal(cfg.get("base"))
    step_dec = _to_decimal(cfg.get("step"))
    base = base_dec if base_dec is not None else Decimal(0)
    step = step_dec if step_dec is not None else Decimal(1)

    def calculator(row: RowData, rf: str = rank_field, b: Decimal = base, s: Decimal = step) -> FieldValue:
        raw_rank = row.get(rf)
        if raw_rank is None:
            return None
        try:
            rank_val = int(raw_rank)
        except Exception as exc:
            msg = "score_by_rank requires integer rank, got {} for rank_field={!r}".format(type(raw_rank).__name__, rf)
            raise TypeError(msg) from exc
        return b - (Decimal(rank_val - 1) * s)

    fingerprint = "rank_field={},base={},step={}".format(rank_field, str(base), str(step))
    return PostFieldSpec(
        out_field_id=str(out_field_id),
        kind="score_by_rank",
        dependencies=(rank_field,),
        fingerprint=fingerprint,
        calculator=calculator,
    )


def _compile_compute_post_field(
    *,
    out_field_id: str,
    cfg: Dict[str, Any],
    engine: SecureComputeEngine,
) -> PostFieldSpec:
    expr = str(cfg.get("expression") or "").strip()
    deps = tuple(str(x) for x in cast("Tuple[str, ...]", cfg.get("dependencies") or ()))  # pragma: allow-cast yaml tuple typed narrowing
    if not expr:
        msg = "aggregate.fields.{} has invalid compute config: missing expression".format(out_field_id)
        raise ValueError(msg)

    try:
        raw_calculator = cast(
            "Callable[..., object]", engine.compile(expr, deps)
        )  # pragma: allow-cast compute engine compile typed narrowing
    except (ScalimComputeExpressionError, ScalimSecurityError) as exc:
        msg = "aggregate.fields.{} has invalid compute expression: {}".format(out_field_id, exc)
        raise ValueError(msg) from exc

    dep_keys = deps

    def calculator(row: RowData, c: Callable[..., object] = raw_calculator) -> FieldValue:
        values = [row.get(key) for key in dep_keys]
        result = c(*values)
        return _ensure_field_value(result, field_id=str(out_field_id), producer="compute")

    return PostFieldSpec(
        out_field_id=str(out_field_id),
        kind="compute",
        dependencies=deps,
        fingerprint=expr,
        calculator=calculator,
    )


def _get_field_name(field_id: str, demand_ir: DemandIr) -> str:
    field_ir = demand_ir.fields.get(field_id)
    name = field_ir.name if field_ir is not None else ""
    if name and name != field_id:
        return str(name)
    return field_id


def _get_derived_field_name(field_id: str, demand_ir: DemandIr, agg: OutputAggregateConfig) -> str:
    field_ir = demand_ir.fields.get(field_id)
    if field_ir is not None:
        return _get_field_name(field_id, demand_ir)

    agg_field = agg.fields.get(field_id)
    if agg_field is not None:
        name = str(agg_field.name or "").strip()
        if name and name != field_id:
            return name
    return field_id


def _export_layout_for_derived(
    *,
    demand_ir: DemandIr,
    agg: OutputAggregateConfig,
    field_ids: Sequence[str],
    header_fields_output_by: str,
) -> ExportLayout:
    normalized = tuple(str(x) for x in field_ids)
    if header_fields_output_by != "name":
        return ExportLayout(field_ids=normalized, header_names=None)

    names: List[str] = []
    has_diff = False
    for fid in normalized:
        resolved = _get_derived_field_name(fid, demand_ir, agg)
        if resolved != fid:
            has_diff = True
        names.append(resolved)
    if not has_diff:
        return ExportLayout(field_ids=normalized, header_names=None)
    return ExportLayout(field_ids=normalized, header_names=tuple(names))


def _output_spec_for_file_resource(file_cfg: FileConfig, *, path: Optional[str], include_header: bool) -> OutputSpec:
    return OutputSpec(
        format="csv",
        path=path,
        encoding=str(file_cfg.encoding),
        streaming=True,
        include_header=bool(include_header),
        sheet_name=None,
    )


def _compile_where_predicate(
    *,
    engine: SecureComputeEngine,
    expression: str,
    requires: Tuple[str, ...],
) -> OutputRowPredicate:
    calc = engine.compile(str(expression), tuple(str(x) for x in requires))
    dep_keys = tuple(str(x) for x in requires)

    def _predicate(row: RowData) -> bool:
        values = [row.get(key) for key in dep_keys]
        return bool(calc(*values))

    return _predicate


def _metric_spec_from_agg_field(*, out_field_id: str, producer_key: str, cfg: Dict[str, Any]) -> AggMetricSpec:
    field_id = str(cfg.get("field")) if cfg.get("field") else None
    field_ids = cfg.get("fields")
    field_ids_norm = tuple(str(x) for x in field_ids) if field_ids else None
    threshold = cfg.get("threshold")
    return AggMetricSpec(
        out_field_id=str(out_field_id),
        op=str(producer_key),
        field_id=field_id,
        field_ids=field_ids_norm,
        threshold=threshold,
    )


def _derived_group_by_spec_from_yaml(
    cfg: OutputAggregateConfig,
    *,
    resolver: SecurePythonReferenceResolver,
    compute_engine: SecureComputeEngine,
) -> DerivedGroupBySpec:
    metric_ids = sorted([fid for fid, fc in cfg.fields.items() if str(fc.producer_key) in _AGG_FUNC_KEYS])
    metrics = tuple(
        _metric_spec_from_agg_field(
            out_field_id=metric_id,
            producer_key=str(cfg.fields[metric_id].producer_key),
            cfg=cast("Dict[str, Any]", cfg.fields[metric_id].config),  # pragma: allow-cast yaml mapping typed narrowing
        )
        for metric_id in metric_ids
    )

    rank_specs: List[RankFieldSpec] = []
    for out_field_id, field_cfg in cfg.fields.items():
        if str(field_cfg.producer_key) not in _RANK_FUNC_KEYS:
            continue
        raw = cast("Dict[str, Any]", field_cfg.config)  # pragma: allow-cast yaml mapping typed narrowing
        partition_by_raw = cast("Tuple[str, ...]", raw.get("partition_by") or ())  # pragma: allow-cast yaml tuple typed narrowing
        order_by_raw = cast("Tuple[str, ...]", raw.get("order_by") or ())  # pragma: allow-cast yaml tuple typed narrowing
        rank_specs.append(
            RankFieldSpec(
                out_field_id=str(out_field_id),
                kind=str(field_cfg.producer_key),
                by=str(raw.get("by")),
                partition_by=tuple(str(x) for x in partition_by_raw),
                order=str(raw.get("order") or "desc"),
                order_by=tuple(str(x) for x in order_by_raw),
                top_k=int(raw.get("top_k") or 0),
                top_k_mode=str(raw.get("top_k_mode") or "rank"),
            )
        )

    post_specs: List[PostFieldSpec] = []
    for out_field_id, field_cfg in cfg.fields.items():
        producer_key = str(field_cfg.producer_key)
        if producer_key == "call_by":
            post_specs.append(
                _compile_call_by_post_field(
                    out_field_id=str(out_field_id),
                    call_by=str(field_cfg.config),
                    resolver=resolver,
                )
            )
        elif producer_key == "score_by_rank":
            post_specs.append(
                _compile_score_by_rank_post_field(
                    out_field_id=str(out_field_id),
                    cfg=cast("Dict[str, Any]", field_cfg.config),  # pragma: allow-cast yaml mapping typed narrowing
                )
            )
        elif producer_key == "compute":
            post_specs.append(
                _compile_compute_post_field(
                    out_field_id=str(out_field_id),
                    cfg=cast("Dict[str, Any]", field_cfg.config),  # pragma: allow-cast yaml mapping typed narrowing
                    engine=compute_engine,
                )
            )

    return DerivedGroupBySpec(
        group_by=tuple(str(x) for x in cfg.group_by),
        metrics=metrics,
        rank_fields=tuple(sorted(rank_specs, key=lambda r: str(r.out_field_id))),
        post_fields=tuple(sorted(post_specs, key=lambda p: str(p.out_field_id))),
        max_groups=int(cfg.max_groups),
        max_distinct=int(cfg.max_distinct),
        distinct_on_overflow=str(cfg.distinct_on_overflow),
    )


def _derived_output_layout_fields(cfg: OutputAggregateConfig) -> Tuple[str, ...]:
    metric_ids = sorted([fid for fid, fc in cfg.fields.items() if str(fc.producer_key) in _AGG_FUNC_KEYS])
    rank_ids = sorted([fid for fid, fc in cfg.fields.items() if str(fc.producer_key) in _RANK_FUNC_KEYS])
    post_ids = sorted([fid for fid, fc in cfg.fields.items() if str(fc.producer_key) in _POST_FUNC_KEYS])
    fields: List[str] = list(cfg.group_by) + metric_ids + rank_ids + post_ids
    return tuple(str(x) for x in fields)


def _compile_extra_sheet(
    *,
    target_id: str,
    cfg: OutputExtraSheetConfig,
    default_sheet: str,
    default_workbook_path: Optional[str],
    default_allow_formulas: bool,
    default_write_lock: bool,
    as_in_memory_csv: bool,
) -> Tuple[OutputSpec, str]:
    sheet = str(cfg.sheet) if cfg.sheet else str(default_sheet)

    if as_in_memory_csv:
        if cfg.path:
            msg = "{}.path is not supported in workflow-managed mode (meta/audit must be written via books write nodes)".format(target_id)
            raise ValueError(msg)
        return (OutputSpec(format="csv", path=None, streaming=True, include_header=True), sheet)

    path = str(cfg.path).strip() if cfg.path else (str(default_workbook_path) if default_workbook_path else "")
    if not path:
        msg = "{} requires a workbook path (set {}.path or provide at least one workbook output)".format(target_id, target_id)
        raise ValueError(msg)
    allow_formulas = default_allow_formulas if cfg.allow_formulas is None else bool(cfg.allow_formulas)
    write_lock = default_write_lock if cfg.write_lock is None else bool(cfg.write_lock)
    return (
        OutputSpec(
            format="excel",
            path=path,
            excel_allow_formulas=allow_formulas,
            write_lock=write_lock,
        ),
        sheet,
    )


def _maybe_compile_extra_sheet(
    *,
    target_id: str,
    cfg: Optional[OutputExtraSheetConfig],
    default_sheet: str,
    default_workbook_path: Optional[str],
    default_allow_formulas: bool,
    default_write_lock: bool,
    skip_without_workbook: bool,
    as_in_memory_csv: bool,
) -> Optional[Tuple[OutputSpec, str]]:
    if cfg is None or (not as_in_memory_csv and skip_without_workbook and cfg.path is None and default_workbook_path is None):
        return None

    out, sheet_name = _compile_extra_sheet(
        target_id=target_id,
        cfg=cfg,
        default_sheet=default_sheet,
        default_workbook_path=default_workbook_path,
        default_allow_formulas=default_allow_formulas,
        default_write_lock=default_write_lock,
        as_in_memory_csv=as_in_memory_csv,
    )
    return out, sheet_name


def _validate_extra_sheet_target_names(config: DemandConfig, *, outputs_path: str) -> None:
    outputs = config.outputs or ()
    reserved = {str(t.name) for t in outputs}
    if config.meta is not None and "meta" in reserved:
        msg = "{}.*.name cannot be 'meta' when meta sheet is enabled".format(outputs_path)
        raise ValueError(msg)
    if config.audit is not None and "audit" in reserved:
        msg = "{}.*.name cannot be 'audit' when audit sheet is enabled".format(outputs_path)
        raise ValueError(msg)


def _effective_file_id_for_output(
    out_cfg: OutputTargetConfig,
    *,
    idx: int,
    outputs_path: str,
) -> Tuple[Optional[str], str]:
    file_ref_path = "{}.{}.to.file".format(outputs_path, int(idx))
    to_cfg = out_cfg.to
    if to_cfg is not None and to_cfg.file is not None:
        candidate = str(to_cfg.file or "").strip()
        if candidate:
            return candidate, file_ref_path
    return None, file_ref_path


def _effective_book_id_for_output(
    out_cfg: OutputTargetConfig,
    *,
    idx: int,
    outputs_path: str,
) -> Tuple[Optional[str], str]:
    book_ref_path = "{}.{}.to.book".format(outputs_path, int(idx))
    to_cfg = out_cfg.to
    if to_cfg is not None and to_cfg.book is not None:
        candidate = str(to_cfg.book or "").strip()
        if candidate:
            return candidate, book_ref_path

    return None, book_ref_path


def _effective_sheet_name_for_output(out_cfg: OutputTargetConfig, *, idx: int, outputs_path: str) -> Tuple[str, str, bool]:
    to_cfg = out_cfg.to
    if to_cfg is not None and to_cfg.sheet is not None:
        sheet_raw = str(to_cfg.sheet or "").strip()
        return sheet_raw, "{}.{}.to.sheet".format(outputs_path, int(idx)), False
    return str(out_cfg.name or ""), "{}.{}.name".format(outputs_path, int(idx)), True


def _require_file_resource(config: DemandConfig, *, file_id: str, file_ref_path: str) -> FileConfig:
    resources = config.resources
    files = resources.files if resources is not None else {}
    file_cfg = files.get(str(file_id))
    if file_cfg is None:
        msg = (
            "Missing file resource id {!r} referenced by {}. "
            "Hint: declare resources.files.{} in the demand YAML, declare workflow.resources.files.{} in the workflow YAML, "
            "or provide overrides.resources.files.{} in Python."
        ).format(str(file_id), str(file_ref_path), str(file_id), str(file_id), str(file_id))
        err = "{} (path={})".format(msg, str(file_ref_path))
        raise ValueError(err)
    return file_cfg


def _require_book_resource(config: DemandConfig, *, book_id: str, book_ref_path: str) -> BookConfig:
    resources = config.resources
    books = resources.books if resources is not None else {}
    book = books.get(str(book_id))
    if book is None:
        msg = (
            "Missing book resource id {!r} referenced by {}. "
            "Hint: declare resources.books.{} in the demand YAML, declare workflow.resources.books.{} in the workflow YAML, "
            "or provide overrides.resources.books.{} in Python."
        ).format(str(book_id), str(book_ref_path), str(book_id), str(book_id), str(book_id))
        err = "{} (path={})".format(msg, str(book_ref_path))
        raise ValueError(err)
    return book


def _resolve_file_export_path(
    config: DemandConfig,
    *,
    file_id: str,
    file_ref_path: str,
    yaml_base_dir: str,
    init_vars: Optional[Dict[str, object]],
) -> Tuple[str, FileConfig]:
    file_cfg = _require_file_resource(config, file_id=str(file_id), file_ref_path=str(file_ref_path))
    export_path = resolve_yaml_relative_output_path(
        file_cfg.path,
        base_dir=str(yaml_base_dir),
        init_vars=init_vars,
        path="resources.files.{}.path".format(str(file_id)),
    )
    return export_path, file_cfg


def _resolve_book_export_path(
    config: DemandConfig,
    *,
    book_id: str,
    book_ref_path: str,
    yaml_base_dir: str,
    init_vars: Optional[Dict[str, object]],
) -> Tuple[str, bool, bool]:
    book = _require_book_resource(config, book_id=str(book_id), book_ref_path=str(book_ref_path))

    kind = str(book.kind or "").strip()
    if kind == "xlsx_file":
        export_path = resolve_yaml_relative_output_path(
            book.path,
            base_dir=str(yaml_base_dir),
            init_vars=init_vars,
            path="resources.books.{}.path".format(str(book_id)),
        )
        return export_path, bool(book.allow_formulas), bool(book.write_lock)

    if kind == "xlsx_memory":
        export = book.export_xlsx
        if export is None:
            msg = (
                "books.kind=xlsx_memory requires export_xlsx for standalone xlsx export (book_id={!r}); "
                "set resources.books.{}.export_xlsx.path or run in a workflow"
            ).format(str(book_id), str(book_id))
            path_ref = "resources.books.{}.export_xlsx".format(str(book_id))
            err = "{} (path={})".format(msg, path_ref)
            raise ValueError(err)
        export_path = resolve_yaml_relative_output_path(
            export.path,
            base_dir=str(yaml_base_dir),
            init_vars=init_vars,
            path="resources.books.{}.export_xlsx.path".format(str(book_id)),
        )
        return export_path, bool(export.allow_formulas), bool(export.write_lock)

    msg = "Unknown book kind: {!r}".format(kind)
    path_ref = "resources.books.{}.kind".format(str(book_id))
    err = "{} (path={})".format(msg, path_ref)
    raise ValueError(err)


def _effective_book_write_defaults(book: BookConfig, *, out_cfg: OutputTargetConfig) -> BookWriteDefaultsConfig:
    base = book.write_defaults
    if base is None:
        base = BookWriteDefaultsConfig(
            mode=str(DEFAULT_BOOK_WRITE_MODE),
            header_policy=str(DEFAULT_BOOK_WRITE_HEADER_POLICY),
        )

    write_cfg = out_cfg.write
    if write_cfg is None:
        return base

    return BookWriteDefaultsConfig(
        mode=str(write_cfg.mode or base.mode or DEFAULT_BOOK_WRITE_MODE),
        align_by=str(write_cfg.align_by or base.align_by),
        header_policy=str(write_cfg.header_policy or base.header_policy or DEFAULT_BOOK_WRITE_HEADER_POLICY),
        on_mismatch=str(write_cfg.on_mismatch or base.on_mismatch),
        on_conflict=str(write_cfg.on_conflict or base.on_conflict),
    )


def _effective_output_header_fields_output_by(
    *,
    out_cfg: OutputTargetConfig,
) -> str:
    write_cfg = out_cfg.write
    if write_cfg is not None and write_cfg.header_fields_output_by is not None:
        return str(write_cfg.header_fields_output_by)

    return str(DEFAULT_OUTPUT_HEADER_BY)


def _effective_output_include_header(
    *,
    out_cfg: OutputTargetConfig,
    mode: Optional[str],
    header_policy: Optional[str],
    include_header_path: str,
) -> bool:
    write_cfg = out_cfg.write
    if mode == "append":
        if write_cfg is not None and write_cfg.include_header is not None:
            msg = "{} is not allowed for append-mode book outputs; use write.header_policy".format(include_header_path)
            raise ValueError(msg)
        return str(header_policy or DEFAULT_BOOK_WRITE_HEADER_POLICY) != "never"

    if write_cfg is not None and write_cfg.include_header is not None:
        return bool(write_cfg.include_header)
    return bool(DEFAULT_OUTPUT_INCLUDE_HEADER)


def compile_output_composition_from_yaml(  # noqa: C901, PLR0912, PLR0915
    config: DemandConfig,
    demand_ir: DemandIr,
    *,
    resolver: SecurePythonReferenceResolver,
    init_vars: Optional[Dict[str, object]] = None,
    yaml_base_dir: Optional[str] = None,
    workflow_managed_output_ids: Optional[FrozenSet[str]] = None,
    outputs_path: str = "outputs",
    skip_extra_sheets_without_workbook: bool = False,
) -> Optional[OutputCompositionSpec]:
    outputs = config.outputs
    if not outputs:
        if config.meta is not None or config.audit is not None:
            msg = "meta/audit requires outputs"
            raise ValueError(msg)
        return None

    _validate_extra_sheet_target_names(config, outputs_path=outputs_path)

    engine = SecureComputeEngine()

    direct_targets: List[OutputTargetSpec] = []
    derived_targets: List[DerivedOutputTargetSpec] = []
    workbook_default_path: Optional[str] = None
    workbook_default_allow_formulas = workbook_default_write_lock = False

    for idx, out_cfg in enumerate(outputs):
        is_primary = idx == 0
        requires = tuple(str(x) for x in (out_cfg.requires or ()))
        requires_opt = requires or None

        in_memory = False
        output_spec: OutputSpec
        header_by = str(DEFAULT_OUTPUT_HEADER_BY)

        file_id, file_ref_path = _effective_file_id_for_output(out_cfg, idx=int(idx), outputs_path=str(outputs_path))
        file_id = str(file_id or "").strip()
        if file_id:
            if yaml_base_dir is None:
                msg = "yaml_base_dir is required to resolve resources.files output paths"
                raise ValueError(msg)
            export_path, file_cfg = _resolve_file_export_path(
                config,
                file_id=str(file_id),
                file_ref_path=str(file_ref_path),
                yaml_base_dir=str(yaml_base_dir),
                init_vars=init_vars,
            )
            include_header = _effective_output_include_header(
                out_cfg=out_cfg,
                mode=None,
                header_policy=None,
                include_header_path="{}.{}.write.include_header".format(outputs_path, idx),
            )
            output_spec = _output_spec_for_file_resource(file_cfg, path=export_path, include_header=include_header)
            header_by = _effective_output_header_fields_output_by(out_cfg=out_cfg)
        else:
            book_id, book_ref_path = _effective_book_id_for_output(out_cfg, idx=int(idx), outputs_path=str(outputs_path))
            if book_id is None:
                msg = (
                    "Missing output destination for output {!r}; set {}.{}.to.file or {}.{}.to.book explicitly. "
                    "Reuse the binding with YAML anchors (`_templates`) or `$import` if needed."
                ).format(str(out_cfg.name), str(outputs_path), int(idx), str(outputs_path), int(idx))
                err = "{} (path={})".format(msg, book_ref_path)
                raise ValueError(err)

            sheet_name, sheet_ref_path, defaulted_from_name = _effective_sheet_name_for_output(
                out_cfg, idx=int(idx), outputs_path=str(outputs_path)
            )
            sheet_name = str(sheet_name or "").strip()
            try:
                _validate_excel_sheet_name(sheet_name, path=sheet_ref_path)
            except ValueError as exc:
                if defaulted_from_name:
                    msg = ("Invalid default Excel sheet name for output {!r}; set {}.{}.to.sheet explicitly. {}").format(
                        str(out_cfg.name), str(outputs_path), int(idx), exc
                    )
                    raise ValueError(msg) from None
                raise

            if workflow_managed_output_ids is not None and str(out_cfg.name) in workflow_managed_output_ids:
                in_memory = True
                effective_defaults = _effective_book_write_defaults(
                    _require_book_resource(config, book_id=str(book_id), book_ref_path=str(book_ref_path)), out_cfg=out_cfg
                )
                include_header = _effective_output_include_header(
                    out_cfg=out_cfg,
                    mode=str(effective_defaults.mode or DEFAULT_BOOK_WRITE_MODE),
                    header_policy=str(effective_defaults.header_policy or DEFAULT_BOOK_WRITE_HEADER_POLICY),
                    include_header_path="{}.{}.write.include_header".format(outputs_path, idx),
                )
                output_spec = OutputSpec(format="csv", path=None, streaming=True, include_header=bool(include_header))
            else:
                if yaml_base_dir is None:
                    msg = "yaml_base_dir is required to resolve resources.books output paths"
                    raise ValueError(msg)
                book = _require_book_resource(config, book_id=str(book_id), book_ref_path=str(book_ref_path))
                effective_defaults = _effective_book_write_defaults(book, out_cfg=out_cfg)
                include_header = _effective_output_include_header(
                    out_cfg=out_cfg,
                    mode=str(effective_defaults.mode or DEFAULT_BOOK_WRITE_MODE),
                    header_policy=str(effective_defaults.header_policy or DEFAULT_BOOK_WRITE_HEADER_POLICY),
                    include_header_path="{}.{}.write.include_header".format(outputs_path, idx),
                )
                export_path, allow_formulas, write_lock = _resolve_book_export_path(
                    config,
                    book_id=str(book_id),
                    book_ref_path=str(book_ref_path),
                    yaml_base_dir=str(yaml_base_dir),
                    init_vars=init_vars,
                )
                output_spec = OutputSpec(
                    format="excel",
                    path=str(export_path),
                    streaming=True,
                    include_header=bool(include_header),
                    sheet_name=str(sheet_name),
                    excel_allow_formulas=bool(allow_formulas),
                    write_lock=bool(write_lock),
                )
                if workbook_default_path is None:
                    workbook_default_path = str(export_path)
                    workbook_default_allow_formulas = bool(allow_formulas)
                    workbook_default_write_lock = bool(write_lock)
            header_by = _effective_output_header_fields_output_by(out_cfg=out_cfg)

        predicate = None
        if out_cfg.where:
            predicate = _compile_where_predicate(engine=engine, expression=str(out_cfg.where), requires=requires)

        if out_cfg.aggregate is None:
            fields = out_cfg.fields or ()
            layout = export_layout_from_demand_ir(
                demand_ir,
                list(fields),
                header_fields_output_by=str(header_by),
            )
            direct_targets.append(
                OutputTargetSpec(
                    target_id=str(out_cfg.name),
                    layout=layout,
                    output=output_spec,
                    in_memory=bool(in_memory),
                    predicate=predicate,
                    is_primary=bool(is_primary),
                    requires=requires_opt,
                )
            )
            continue

        agg = out_cfg.aggregate
        derived = _derived_group_by_spec_from_yaml(agg, resolver=resolver, compute_engine=engine)
        out_layout_fields = tuple(str(x) for x in out_cfg.fields) if out_cfg.fields is not None else _derived_output_layout_fields(agg)
        out_layout = _export_layout_for_derived(
            demand_ir=demand_ir,
            agg=agg,
            field_ids=out_layout_fields,
            header_fields_output_by=str(header_by),
        )
        derived_targets.append(
            DerivedOutputTargetSpec(
                target_id=str(out_cfg.name),
                derived=derived,
                output_layout=out_layout,
                output=output_spec,
                in_memory=bool(in_memory),
                predicate=predicate,
                is_primary=bool(is_primary),
                requires=requires_opt,
            )
        )

    meta_sheet_spec = None
    meta_sheet_compiled = _maybe_compile_extra_sheet(
        target_id="meta",
        cfg=config.meta,
        default_sheet="__meta__",
        default_workbook_path=workbook_default_path,
        default_allow_formulas=workbook_default_allow_formulas,
        default_write_lock=workbook_default_write_lock,
        skip_without_workbook=skip_extra_sheets_without_workbook,
        as_in_memory_csv=workflow_managed_output_ids is not None,
    )
    if meta_sheet_compiled is not None:
        meta_out, meta_sheet_name = meta_sheet_compiled
        meta_sheet_spec = MetaSheetSpec(
            target_id="meta",
            output=meta_out,
            sheet_name=meta_sheet_name,
            in_memory=workflow_managed_output_ids is not None,
        )

    audit_sheet_spec = None
    audit_sheet_compiled = _maybe_compile_extra_sheet(
        target_id="audit",
        cfg=config.audit,
        default_sheet="__audit__",
        default_workbook_path=workbook_default_path,
        default_allow_formulas=workbook_default_allow_formulas,
        default_write_lock=workbook_default_write_lock,
        skip_without_workbook=skip_extra_sheets_without_workbook,
        as_in_memory_csv=workflow_managed_output_ids is not None,
    )
    if audit_sheet_compiled is not None:
        audit_out, audit_sheet_name = audit_sheet_compiled
        audit_sheet_spec = AuditSheetSpec(
            target_id="audit",
            output=audit_out,
            sheet_name=audit_sheet_name,
            in_memory=workflow_managed_output_ids is not None,
        )

    return OutputCompositionSpec(
        targets=tuple(direct_targets),
        derived_targets=tuple(derived_targets),
        meta_sheet=meta_sheet_spec,
        audit_sheet=audit_sheet_spec,
        failure_policy=str(config.failure_policy or "all_fail"),
        include_full_error_message=bool(config.include_full_error_message),
    )


__all__ = [
    "compile_output_composition_from_yaml",
]



from typing import TYPE_CHECKING, Callable, Dict, Hashable, List, Optional, Sequence, Set, Tuple, cast

if TYPE_CHECKING:
    from concurrent.futures import Future

from ....planning.operators import LoadRefOperatorIr
from ...context import BatchContext
from ...executor.operators.load_ref.executor import LoadRefOperatorExecutor
from ...executor.runtime.runtime import ExecutionRuntime
from ..capture import HookCaptureManager
from ..overlay_context import OverlayBatchContext
from ..strategy_unit import TaskSpec as _TaskSpec
from ..submission_unit import LayerScheduleStats as _LayerScheduleStats
from ..submission_unit import run_tasks_in_pool as _run_tasks_in_pool_unit
from .loadref_scheduler_base import AdaptiveLoadRefSchedulerBase
from .loadref_scheduler_support import AdaptiveTaskResult as _AdaptiveTaskResult


class AdaptiveLoadRefSchedulerExecutionMixin(AdaptiveLoadRefSchedulerBase):
    def _build_loadref_executor(self) -> LoadRefOperatorExecutor:
        factory = self._require_overrides().adaptive_loadref_executor_factory
        if factory is not None:
            return cast("LoadRefOperatorExecutor", factory())  # pragma: allow-cast overrides factory typed narrowing
        return LoadRefOperatorExecutor()

    def _execute_ops_serially(
        self,
        ops: Sequence[LoadRefOperatorIr],
        *,
        context: BatchContext,
        batch_row_nth: List[Hashable],
        runtime: ExecutionRuntime,
        serial_executor: LoadRefOperatorExecutor,
        after_operator: Optional[Callable[[LoadRefOperatorIr], None]],
    ) -> None:
        for op in ops:
            serial_executor.execute(op, context, batch_row_nth, runtime)
            if after_operator is not None:
                after_operator(op)

    def _run_task(
        self,
        spec: _TaskSpec,
        base_context: BatchContext,
        batch_row_nth: List[Hashable],
        runtime: ExecutionRuntime,
        required_fields: Optional[Set[str]],
    ) -> _AdaptiveTaskResult:
        hook_manager = HookCaptureManager(runtime.hook_manager)
        observer_manager = runtime.observer_manager.create_capture_manager()

        task_runtime = ExecutionRuntime(
            plan=self._require_plan(),
            hook_manager=hook_manager,
            observer_manager=observer_manager,
            main_source=runtime.main_source,
            guardrails=runtime.guardrails,
            loader_retry=runtime.loader_retry,
            parallel_mode="seq",
            max_workers=0,
            key_normalization=runtime.key_normalization,
        )
        task_runtime.preloaded_cache = runtime.preloaded_cache
        task_runtime.batch_num = runtime.batch_num

        task_context = OverlayBatchContext(base_context, required_fields=required_fields)
        self._build_loadref_executor().execute(spec.op, task_context, batch_row_nth, task_runtime)

        overlay = task_context.drain_overlay()
        hook_events = hook_manager.drain_events()
        observer_events = observer_manager.drain_events()
        return _AdaptiveTaskResult(
            overlay=overlay,
            hook_events=hook_events,
            observer_events=observer_events,
            relation_key=spec.relation_key,
            group_enabled=spec.group_enabled,
        )

    def _run_tasks_in_pool(
        self,
        task_order: Sequence[Tuple[str, object]],
        task_specs: Dict[Tuple[str, object], _TaskSpec],
        *,
        max_workers: int,
        submit_task: Callable[[_TaskSpec], "Future[_AdaptiveTaskResult]"],
        collect_stats: bool,
    ) -> Tuple[Dict[Tuple[str, object], _AdaptiveTaskResult], Optional[_LayerScheduleStats]]:
        results_by_key, layer_stats = _run_tasks_in_pool_unit(
            task_order,
            task_specs,
            max_workers=max_workers,
            submit_task=submit_task,
            collect_stats=collect_stats,
            resolve_pool_limit=lambda pool_name, resolved: self._require_tuning().resolve_pool_limit(
                pool_name,
                resolved_max_workers=resolved,
            ),
        )
        return results_by_key, layer_stats


__all__ = []



# region imports
import json
import logging
import time
from collections.abc import Sized
from typing import Any, Dict, Hashable, Iterable, List, Optional, Union, cast

from ...events._events import (
    BatchEndEvent,
    BatchStartEvent,
    FieldSlimEvent,
    LoaderCallEvent,
    PipelineEndEvent,
    PipelineStartEvent,
    RowReleaseEvent,
    RowWriteEvent,
)
from ...vendor.dataclassesx import asdict, dataclass, field
from .._internal.console_report import emit_info, format_seconds
from ..observer import EventDispatchObserver

# endregion

_LOGGER = logging.getLogger(__name__)


@dataclass
class LoaderCallStep:
    """加载器调用步骤元信息"""

    step_type: str = "loader_call"
    loader_name: str = ""
    params: Dict[str, str] = field(default_factory=dict)
    result_count: int = 0
    duration: float = 0.0
    timestamp: float = field(default_factory=time.time)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def serialize_params(params: Dict[str, Any]) -> Dict[str, str]:
        result: Dict[str, str] = {}
        for key, pv in params.items():
            if isinstance(pv, (set, list, tuple)):
                items = list(cast("Iterable[Any]", pv))  # pragma: allow-cast iterable typed narrowing
                result[key] = str(items)
            else:
                result[key] = str(pv)
        return result


@dataclass
class FieldSlimStep:
    """字段瘦身步骤"""

    step_type: str = "field_slim"
    field_key: str = ""
    reason: str = ""
    remaining_fields: Optional[int] = None
    timestamp: float = field(default_factory=time.time)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class RowWriteStep:
    """行写入步骤"""

    step_type: str = "row_write"
    row_id: Optional[Hashable] = None
    batch_num: Optional[int] = None
    timestamp: float = field(default_factory=time.time)

    def to_dict(self) -> Dict[str, Any]:
        result = asdict(self)
        if self.row_id is not None:
            result["row_id"] = str(self.row_id)
        return result


@dataclass
class BatchTrace:
    """单个批次的执行追踪"""

    batch_num: Optional[int] = None
    row_ids: List[Hashable] = field(default_factory=list)
    steps: "List[Union[LoaderCallStep, FieldSlimStep, RowWriteStep]]" = field(default_factory=list)
    start_time: float = field(default_factory=time.time)
    end_time: Optional[float] = None
    duration: Optional[float] = None

    def add_step(self, step: "Union[LoaderCallStep, FieldSlimStep, RowWriteStep]") -> None:
        self.steps.append(step)

    def finish(self) -> None:
        self.end_time = time.time()
        self.duration = self.end_time - self.start_time

    def to_dict(self) -> Dict[str, Any]:
        result = asdict(self)
        result["row_ids"] = [str(k) for k in self.row_ids]
        return result


class ExecutionTraceObserver(EventDispatchObserver):
    """执行追踪观察者:记录详细的执行步骤"""

    def __init__(self) -> None:
        self.batches: List[BatchTrace] = []
        self.current_batch: Optional[BatchTrace] = None

        self.pipeline_start_time: Optional[float] = None
        self.pipeline_end_time: Optional[float] = None
        self.target_fields: List[str] = []
        self.batch_size: Optional[int] = None

        self.total_loader_calls: int = 0
        self.total_field_slims: int = 0
        self.total_row_writes: int = 0

    def on_pipeline_start(self, event: PipelineStartEvent) -> None:
        self.pipeline_start_time = time.time()
        self.target_fields = event.targets
        self.batch_size = event.batch_size

    def on_pipeline_end(self, event: PipelineEndEvent) -> None:
        _ = event
        self.pipeline_end_time = time.time()

    def on_batch_start(self, event: BatchStartEvent) -> None:
        self.current_batch = BatchTrace(
            batch_num=event.batch_num,
            row_ids=event.row_ids,
            start_time=time.time(),
        )

    def on_batch_end(self, event: BatchEndEvent) -> None:
        _ = event
        if self.current_batch:
            self.current_batch.finish()
            self.batches.append(self.current_batch)
            self.current_batch = None

    def on_loader_call(self, event: LoaderCallEvent) -> None:
        result_count = len(event.result) if isinstance(event.result, Sized) else 0
        self.total_loader_calls += 1
        if self.current_batch:
            step = LoaderCallStep(
                loader_name=event.loader_name,
                params=LoaderCallStep.serialize_params(event.params),
                result_count=result_count,
                duration=event.duration,
            )
            self.current_batch.add_step(step)

    def on_field_slim(self, event: FieldSlimEvent) -> None:
        if self.current_batch:
            step = FieldSlimStep(
                field_key=event.field_key,
                reason=event.reason,
                remaining_fields=event.remaining_fields,
            )
            self.current_batch.add_step(step)
            self.total_field_slims += 1

    def on_row_write(self, event: RowWriteEvent) -> None:
        if self.current_batch:
            step = RowWriteStep(
                row_id=event.row_id,
                batch_num=event.batch_num,
            )
            self.current_batch.add_step(step)
            self.total_row_writes += 1

    def on_row_release(self, event: RowReleaseEvent) -> None:
        _ = event

    def export_to_json(self, indent: int = 2) -> str:
        data = {
            "pipeline": {
                "start_time": self.pipeline_start_time,
                "end_time": self.pipeline_end_time,
                "batch_size": self.batch_size,
                "target_fields": self.target_fields,
            },
            "stats": {
                "total_batches": len(self.batches),
                "total_loader_calls": self.total_loader_calls,
                "total_field_slims": self.total_field_slims,
                "total_row_writes": self.total_row_writes,
            },
            "batches": [batch.to_dict() for batch in self.batches],
        }
        return json.dumps(data, ensure_ascii=False, indent=indent, default=str)

    def print_summary(self) -> None:
        emit_info(
            _LOGGER,
            "execution_trace",
            "summary",
            total_batches=len(self.batches),
            total_loader_calls=int(self.total_loader_calls),
            total_field_slims=int(self.total_field_slims),
            total_row_writes=int(self.total_row_writes),
        )

        if not self.batches:
            return

        last_batch = self.batches[-1]
        loader_steps = [s for s in last_batch.steps if isinstance(s, LoaderCallStep)]
        slim_count = sum(1 for s in last_batch.steps if isinstance(s, FieldSlimStep))
        write_count = sum(1 for s in last_batch.steps if isinstance(s, RowWriteStep))
        row_ids = [str(k) for k in last_batch.row_ids]
        row_id_samples = ",".join(row_ids[:3]) if row_ids else None

        emit_info(
            _LOGGER,
            "execution_trace",
            "last_batch",
            batch_num=int(last_batch.batch_num) if last_batch.batch_num is not None else None,
            row_count=len(row_ids),
            row_id_samples=row_id_samples,
            duration_s=format_seconds(last_batch.duration, digits=2),
            steps=len(last_batch.steps),
            loader_calls=len(loader_steps),
            field_slims=int(slim_count),
            row_writes=int(write_count),
        )

        if loader_steps:
            total = len(loader_steps)
            showing = min(5, total)
            emit_info(_LOGGER, "execution_trace", "loader_call_samples", total=total, showing=int(showing))
            for idx, step in enumerate(loader_steps[:showing], 1):
                emit_info(
                    _LOGGER,
                    "execution_trace",
                    "loader_call",
                    idx=int(idx),
                    loader=str(step.loader_name),
                    records=int(step.result_count),
                    duration_s=format_seconds(float(step.duration), digits=2),
                )


__all__ = [
    "ExecutionTraceObserver",
    "FieldSlimStep",
    "LoaderCallStep",
    "RowWriteStep",
]



from typing import Optional, Sequence

from .runtime.introspection import OutputConfigDict
from .runtime.introspection import load_output_config as _load_output_config
from .runtime.references import derive_base_module_path as _derive_base_module_path


def load_output_config(yaml_path: str) -> OutputConfigDict:
    return _load_output_config(yaml_path)


def derive_base_module_path(
    yaml_path: str,
    *,
    sys_path: Optional[Sequence[Optional[str]]] = None,
    cwd: Optional[str] = None,
) -> str:
    return _derive_base_module_path(yaml_path, sys_path=sys_path, cwd=cwd)


__all__ = [
    "OutputConfigDict",
    "derive_base_module_path",
    "load_output_config",
]
